"""Functions for determining and setting slab symmetry."""

__authors__ = (
    'Florian Kraushofer (@fkraushofer)',
    )
__copyright__ = 'Copyright (c) 2019-2024 ViPErLEED developers'
__created__ = '2020-08-19'
__license__ = 'GPLv3+'

import copy
import itertools
import logging
import re

import numpy as np
import scipy.spatial as sps

from viperleed.calc.classes.atom_containers import AtomList
from viperleed.calc.classes.slab import AlreadyMinimalError
from viperleed.calc.classes.sym_entity import SymPlane
from viperleed.calc.files import parameters
from viperleed.calc.lib import leedbase
from viperleed.calc.lib.base import addUnequalPoints
from viperleed.calc.lib.base import angle
from viperleed.calc.lib.base import dist_from_line
from viperleed.calc.lib.base import rotation_matrix
from viperleed.calc.lib.base import rotation_matrix_order

logger = logging.getLogger(__name__)


def getSymPosLists(sl, rp, pointlist, output=False):
    """Generates and returns a symposlist and hexsymposlist based on the
    pointlist, for example the list of cartesian in-plane atom positions
    in the lowest-occupied layer."""

    def uniqueSymPosList(sl, rp, spl, verbose, description=""):
        eps = rp.SYMMETRY_EPS
        abst = np.transpose(sl.ab_cell)
        tree = sps.KDTree(spl)
        if len(spl) > 1e6:
            logger.warning(
                "Approximate search for symmetry positions will "
                "be applied due to very large number of candidates - check "
                "result for errors! If the search fails, consider lowering "
                "the SYMMETRY_EPS z component or setting the "
                "SYMMETRY_FIND_ORI parameter to False.")
            rp.setHaltingLevel(2)
            approximate = min([np.linalg.norm(abst[0]),
                               np.linalg.norm(abst[1])])/100
            # significantly speeds up the search especially for
            #   large unit cells, but might give some errors.
        else:
            approximate = 0
        usepoint = [True]*len(spl)
        if verbose:
            t = int(len(spl)/10.0)
        for (i, p) in enumerate(spl):
            if verbose:
                if (i+1) % t == 0:
                    logger.debug(description+": "
                                 + str(int((i+1)/t)*10)+"%")
            if usepoint[i]:
                for j in tree.query_ball_point(p, eps,
                                               eps=approximate)[1:]:
                    usepoint[j] = False
        spl = list(itertools.compress(spl, usepoint))
        return(spl)

    symposlist = [np.array([0., 0.])]    # always check the origin
    hexsymposlist = []
    symposlist.extend(pointlist)
    symposlist.extend([(p1+p2)/2 for (p1, p2) in
                       itertools.combinations(pointlist, 2)])
    if sl.celltype == "hexagonal":
        hexsymposlist = [(p1+p2+p3)/3 for (p1, p2, p3) in
                         itertools.combinations(pointlist, 3)]
    # collapse to base unit cell:
    symposlist = list(np.dot(sl.ab_cell,
                             (np.dot(np.linalg.inv(sl.ab_cell),
                                     np.array(symposlist).transpose())
                              % 1.0)).T)
    if len(hexsymposlist) > 0:
        hexsymposlist = list(np.dot(sl.ab_cell,
                                    (np.dot(np.linalg.inv(sl.ab_cell),
                                            np.array(hexsymposlist).T)
                                     % 1.0)).T)
    # remove duplicates:
    verbose = (False if (len(symposlist)+len(hexsymposlist) < 1e5
                         or not output) else True)
    if verbose:
        logger.debug("Found {} candidates, removing duplicates..."
                     .format(len(symposlist)+len(hexsymposlist)))
    # symposlist:
    symposlist = uniqueSymPosList(sl, rp, symposlist, verbose,
                                  description="Atoms and pairs")
    # hexsymposlist:
    if len(hexsymposlist) > 1:
        hexsymposlist = uniqueSymPosList(sl, rp, hexsymposlist, verbose,
                                         description="Triplets")
    return(symposlist, hexsymposlist)


def findBulkSymmetry(sl, rp):
    """Checks the bulk slab for screw axes and glide planes."""
    eps = rp.SYMMETRY_EPS
    abt = sl.ab_cell.T.copy()
    rotsfound = []
    glidesfound = []
    ts = copy.deepcopy(sl)
    rp2 = copy.deepcopy(rp)
    # optimize C vector, removing vacuum below first. Notice that
    # this is the safer way, as it allows us to relax assumptions
    # on the z periodicity of sl
    ts.remove_vacuum_at_bottom(rp2)
    try:
        ts.ensure_minimal_c_vector(rp2)
    except AlreadyMinimalError:
        pass
    else:
        logger.debug('Bulk unit cell could be reduced with repeat vector '
                     '[{:.5f} {:.5f} {:.5f}]'.format(*(rp2.BULK_REPEAT)))

    # figure out what to check
    pcands = ts.get_candidate_layer_periods(eps)
    if len(pcands) == 0:
        return
    nl = ts.n_sublayers
    # check for screw axes
    checkrots = []
    if nl % 2 == 0:
        checkrots.extend([2, 4])
    if ts.celltype == "hexagonal" and (nl % 3 == 0):
        checkrots.extend([3, 6])
    for per in pcands:
        for ro in [ro for ro in checkrots if ro not in rotsfound]:
            if ts.is_bulk_screw_symmetric(ro, per, eps):
                rotsfound.append(ro)
    sl.bulk_screws = rotsfound
    if len(rotsfound) > 0:
        logger.debug("Bulk screw axes found: " +
                     ", ".join([str(v) for v in rotsfound]))
    # check for glide planes
    ori = np.array([0, 0])
    checkglides = [SymPlane(ori, abt[0], abt), SymPlane(ori, abt[1], abt),
                   SymPlane(ori, abt[0] + abt[1], abt),
                   SymPlane(ori, abt[0] - abt[1], abt)]
    if ts.celltype == "hexagonal":
        checkglides.extend([SymPlane(ori, 2*abt[0] + abt[1], abt),
                            SymPlane(ori, abt[0] + 2*abt[1], abt)])
    for per in pcands:
        for gl in [gl for gl in checkglides if gl not in glidesfound]:
            if ts.is_bulk_glide_symmetric(gl, per, eps):
                glidesfound.append(gl)
    sl.bulk_glides = glidesfound
    if len(rotsfound) > 0:
        logger.debug("Bulk glide planes found: " +
                     ", ".join([str(gl.par) for gl in glidesfound]))
    return


def findSymmetry(sl, rp, bulk=False, output=True, forceFindOri=False):
    """Reduces the unit cell if necessary and finds the plane group of the
    slab. Stores the plane group and the higher-symmetry direction of the
    unit cell, if there is one."""
    celltype = "ERROR - not recognized"
    planegroup = ""  # plane group will be stored in Hermann-Mauguin notation
    eps = rp.SYMMETRY_EPS
    # reduce surface unit cell
    abst = sl.ab_cell.T  # surface unit cell, transposed
#        usurf = np.array([[1,0],[0,1]])
    if rp.SYMMETRY_FIX != "p1":
        abst, usurf, celltype = leedbase.reduceUnitCell(abst)
    else:
        # if symmetry is switched off, don't try to change the cell.
        celltype, usurf = leedbase.checkLattice(abst)
    # usurf tracks unit cell changes
    # reduce bulk unit cell
    if not bulk:
        abbt = np.dot(np.linalg.inv(rp.SUPERLATTICE), abst)
        # bulk ab unit cell, transposed
        abbt, ubulk, _ = leedbase.reduceUnitCell(abbt)
        # ubulk tracks unit cell changes
    utr = np.array([[0, 0, 0], [0, 0, 0], [0, 0, 1]])
    utr[:2, :2] = usurf
    if not np.array_equal(utr, np.identity(3, dtype=int)):
        if (np.array_equal(utr, np.array([[0, -1, 0], [1, 0, 0], [0, 0, 1]]))
                and output):
            logger.info("The POSCAR unit cell was changed from an acute "
                        "to an obtuse form.")
            rp.checklist.append(
                "The POSCAR unit vector definitions have changed. Make sure "
                "to check beam labels for compatibility with new unit cell.")
        elif output:
            logger.warning("The POSCAR unit cell was not in its highest "
                           "symmetry form. The unit cell will be modified "
                           "with the transformation matrix: \n"+str(utr))
            rp.setHaltingLevel(1)
            rp.checklist.append(
                "The POSCAR unit vector definitions have changed. Make sure "
                "to check beam labels for compatibility with new unit cell.")
        # MODIFY SUPERLATTICE PARAMETER
        if not bulk and rp.superlattice_defined:
            rp.SUPERLATTICE = np.dot(usurf, np.dot(rp.SUPERLATTICE,
                                                   np.linalg.inv(ubulk)))
            parameters.modify(rp, "SUPERLATTICE")
        # MODIFY SYMMETRY_FIX PARAMETER
        if "[" in rp.SYMMETRY_FIX and not bulk:
            rgx = re.compile(r'\s*(?P<group>(pm|pg|cm|rcm|pmg))\s*\[\s*'
                             + r'(?P<i1>[-012]+)\s+(?P<i2>[-012]+)\s*\]')
            m = rgx.match(rp.SYMMETRY_FIX)
            targetsym = m.group("group")
            tspar = [int(m.group("i1")), int(m.group("i2"))]
            cartdir = np.dot(tspar, abst)
            newab = np.dot(sl.ab_cell, np.transpose(usurf))
            newdir = np.dot(np.linalg.inv(newab), cartdir)
            newdir = newdir / min(newdir)
            s = (targetsym+"[{:.0f} {:.0f}]".format(newdir[0], newdir[1]))
            parameters.modify(rp, "SYMMETRY_FIX", s)
        # MODIFY UNIT CELL
        sl.update_cartesian_from_fractional()
        sl.ucell_mod.append(('rmul', utr.T))
        sl.ucell = np.dot(sl.ucell, utr.T)
        # same as np.transpose(np.dot(utr,np.transpose(sl.ucell)))
        sl.collapse_cartesian_coordinates(update_origin=True)
        # gets fractional coordinates in the new unit cell and
        #   collapses appropriately

    # check cell type again
    abst = sl.ab_cell.T
    celltype, _ = leedbase.checkLattice(abst)
    sl.celltype = celltype
    if output:
        logger.info("Found unit cell type: "+celltype)
        logger.info("Starting symmetry search...")
    # FIND HIGHEST SYMMETRY ORIGIN
    sl.collapse_cartesian_coordinates()
    # create a testslab: C projected to Z
    ts = copy.deepcopy(sl)
    if bulk:        # check whether there are at least 2 atomic layers
        ts.create_sublayers(eps.z)
        if ts.n_sublayers < 2:
            ts = ts.with_double_thickness()
    ts.project_c_to_z()
    ts.sort_by_z()

    bigslab = copy.deepcopy(ts)
    # will have atoms duplicated and shifted to 4 unit cells
    #  ([0,0], [0,1], [1,0], [1,1])
    bigslab.atlist.strict = False
    for at in bigslab.atlist[:]:
        for i in range(0, 2):
            for j in range(0, 2):
                if not (i == 0 and j == 0):
                    tmpat = at.duplicate()
                    tmpat.pos[0] += i
                    tmpat.pos[1] += j
    bigslab.update_atom_numbers()
    bigslab.atlist.strict = True
    bigslab.update_cartesian_from_fractional(update_origin=True)
    # bigslab.full_update(rp)   can't do this - would collapse coordinates!
    bigslab.create_sublayers(eps.z)

    # find the lowest occupancy sublayer; comparing candidate
    #   axes / planes to this one will be fastest
    lowocclayer = bigslab.fewest_atoms_sublayer
    minlen = lowocclayer.n_atoms

    # find candidate positions for symmetry points / planes:
    if not rp.SYMMETRY_FIND_ORI and not forceFindOri and not bulk:
        symposlist = [np.array([0., 0.])]
        # only check origin, planes are defined explicitly for this case
        hexsymposlist = []
    else:
        if output:
            logger.debug("Generating candidate high-symmetry positions "
                         "from layer with "+str(int(minlen/4))+" atoms...")
        if minlen > 400:
            logger.warning(
                "The number of atoms in the smallest sublayer is very large. "
                "This can make the symmetry search take a very long time. To "
                "avoid this, either decrease the z component of the "
                "SYMMETRY_EPS parameter, or set the SYMMETRY_FIND_ORI "
                "parameter to False.")
        pl = [at.cartpos[0:2] for at in lowocclayer]
        symposlist, hexsymposlist = getSymPosLists(sl, rp, pl, output)

    comsymposlist = addUnequalPoints(symposlist, hexsymposlist, eps,
                                     uniqueLists=True)

    # we're done with the bigger slab, actually testing symmetry operations
    #   can be done just on the basic one.
    ts.create_sublayers(eps.z)
    lowocclayer = ts.sublayers[bigslab.sublayers.index(lowocclayer)]
    del bigslab

    # find potential rotation axes:
    if (rp.SYMMETRY_FIND_ORI or forceFindOri) and output:
        logger.debug("Checking for rotation axes: "
                     + str(len(comsymposlist))+" candidates...")
    # test potential rotation axes:
    toprotsym = 0   # keep track of the highest rotational symmetry so far

    toprotlist = []  # positions of highest rot-symmetry points
    for p in comsymposlist:
        rotsymorder = 0
        if ts.is_rotation_symmetric(p, 2, eps):
            rotsymorder = 2
            if celltype == "square":
                if ts.is_rotation_symmetric(p, 4, eps):
                    rotsymorder = 4
        if celltype == "hexagonal":
            if ts.is_rotation_symmetric(p, 3, eps):
                rotsymorder = 3
                if ts.is_rotation_symmetric(p, 6, eps):
                    rotsymorder = 6
        if rotsymorder > toprotsym:     # new best point found
            toprotsym = rotsymorder
            toprotlist = [p]
        elif rotsymorder != 0 and rotsymorder == toprotsym:
            toprotlist.append(p)
    if toprotsym > 0 and output:
        logger.debug("Highest rotation axis has order " + str(toprotsym))

    if toprotsym == 0 or toprotlist:
        # no rotation symmetry or multiple rotation centers with same order
        if output:
            logger.debug("Checking for mirror/glide planes...")

        # check for mirror/glides
        mirror = False
        glide = False
        symplanelist = []
        test_mirror_dirs = [(1, 0), (0, 1), (1, 1), (1, -1)]
        if celltype == "hexagonal":
            test_mirror_dirs.extend([(1, 2), (2, 1)])
        if not rp.SYMMETRY_FIND_ORI and not forceFindOri and not bulk:
            for (pa, pb) in [(0, 0), (0.25, 0.25), (0.25, -0.25)]:
                for (i, j) in test_mirror_dirs:
                    symplanelist.append(SymPlane(pa*abst[0]+pb*abst[1],
                                                 i*abst[0]+j*abst[1], abst,
                                                 index2=True))
        else:
            if toprotsym == 0:
                checklist = symposlist
            else:
                checklist = toprotlist
            for (k, pos) in enumerate(checklist):
                for (i, j) in test_mirror_dirs:
                    symplanelist.append(SymPlane(pos, i*abst[0]+j*abst[1],
                                                 abst, index2=True))
        if output:
            logger.debug(str(len(symplanelist))
                         + " candidates for mirror/glide planes found...")
        for spl in symplanelist:    # test the candidates
            if ts.is_mirror_symmetric(spl, eps):
                spl.type = "mirror"
                mirror = True
            elif ts.is_mirror_symmetric(spl, eps, glide=True):
                spl.type = "glide"
                glide = True
        droptypes = ["none"]
        if toprotsym > 0 and mirror:
            droptypes.append("glide")
        i = 0
        while i < len(symplanelist):
            if symplanelist[i].type in droptypes:
                symplanelist.pop(i)
            else:
                i += 1

    if toprotsym > 0:
        # identify origin
        if symplanelist:
            # if there are no symmetry planes, then keep the old list
            toprotlist = [spl.pos for spl in symplanelist]
        topsympoint = toprotlist[0]
        corners = (np.zeros(2), abst[0], abst[1], abst[0]+abst[1])
        mindist = min(np.linalg.norm(topsympoint-c) for c in corners)
        for p in toprotlist[1:]:
            if min(np.linalg.norm(p-c) for c in corners) < mindist:
                # for points with equal rotational symmetry, prioritize the
                #   one closest to the origin to avoid shifting the unit cell
                #   randomly every time
                mindist = min(np.linalg.norm(p-c) for c in corners)
                topsympoint = p
        # shift origin
        sl.translate_atoms_2d(-topsympoint)
        ts.translate_atoms_2d(-topsympoint)

    if toprotsym == 0: # should be an else
        # identify group:
        oriplane = None
        if not mirror:
            if not glide:
                planegroup = "p1"   # no origin shift required.
            else:
                planegroup = "pg"
                for spl in symplanelist:
                    if oriplane is None:
                        oriplane = spl
                    elif (spl.distanceFromOrigin(abst) <
                          oriplane.distanceFromOrigin(abst)):
                        # prioritize planes close to the origin of cell (1,1)
                        oriplane = spl
        else:
            if not glide:
                if celltype in ("square", "rectangular"):
                    planegroup = "pm"
                else:
                    planegroup = "p1"
            else:
                planegroup = "cm"
            for spl in symplanelist:
                if spl.type == "mirror":
                    if oriplane is None:
                        oriplane = spl
                    elif (spl.distanceFromOrigin(abst) <
                          oriplane.distanceFromOrigin(abst)):
                        oriplane = spl
                        # prioritize planes close to the origin of cell (1, 1)
            if planegroup == "cm":
                if celltype in ("square", "rectangular"):
                    # both mirrors and glides. if parallel to unit vectors: rcm
                    if tuple(oriplane.par) in [(1, 0), (0, 1)]:
                        planegroup = "rcm"
                elif (celltype == "hexagonal"
                      and tuple(oriplane.par) not in [(1, 1), (1, -1)]):
                    # Special case - requires rotation of unit cell to bring
                    # mirror/glide along either (11) or (1-1) direction.
                    abst, oriplane = mirror_to_diagonal(sl, rp, abst, oriplane)
        if oriplane is not None and oriplane.distanceFromOrigin(abst) > eps:
            # shift to closest point on oriplane
            shiftv = (np.array([oriplane.dir[1], -oriplane.dir[0]])
                      * oriplane.distanceFromOrigin(abst))
            if dist_from_line(
                    oriplane.pos, oriplane.pos+oriplane.dir, shiftv) > eps:
                shiftv = -1*shiftv
            # ts is not used any more in this case, otherwise those atoms
            #  would have to be shifted as well.
            sl.translate_atoms_2d(-shiftv)
        if oriplane:
            oriplane.pos = np.array([0, 0])
            sl.orisymplane = oriplane

    if not planegroup:
        # start by checking special case: in cmm, there are two inequivalent
        #  2fold axes, one of which would not have been found yet -> shift
        #  there (potentially), test
        if toprotsym == 2 and celltype in ["hexagonal", "rhombic"]:
            shiftslab = copy.deepcopy(ts)
            shiftslab.translate_atoms_2d(-abst[0]/2)
            # test diagonal mirror at shifted origin
            spl = SymPlane(np.array([0, 0]), (abst[0]+abst[1]), abst)
            if shiftslab.is_mirror_symmetric(spl, eps):
                planegroup = "cmm"
                ts = shiftslab
                sl.translate_atoms_2d(-abst[0]/2)  # correct origin

    if not planegroup:
        efftype = ""    # effective cell type
        if celltype == "hexagonal":
            if not (toprotsym == 3 or toprotsym == 6):
                efftype = "rhombic"
            else:
                # test mirror plane along unit vector at origin
                spl = SymPlane(np.array([0, 0]), abst[0], abst)
                if ts.is_mirror_symmetric(spl, eps):
                    if toprotsym == 6:
                        planegroup = "p6m"
                    else:
                        planegroup = "p31m"
                else:
                    if toprotsym == 6:
                        planegroup = "p6"
                    else:
                        # test mirror plane along both diagonals at origin
                        found = False
                        for i in [+1, -1]:
                            spl = SymPlane(np.array([0, 0]),
                                           (abst[0]+(i*abst[1])), abst)
                            if ts.is_mirror_symmetric(spl, eps):
                                found = True
                                break
                        if found:
                            planegroup = "p3m1"
                        else:
                            planegroup = "p3"
        elif celltype == "square":
            if toprotsym == 4:
                # test mirror plane along unit vector at origin
                spl = SymPlane(np.array([0, 0]), abst[0], abst)
                if ts.is_mirror_symmetric(spl, eps):
                    planegroup = "p4m"
                else:
                    # test glide plane along diagonal at origin
                    spl = SymPlane(np.array([0, 0]), (abst[0]+abst[1]),
                                   abst)
                    if ts.is_mirror_symmetric(spl, eps, glide=True):
                        planegroup = "p4g"
                    else:
                        planegroup = "p4"
            else:   # p1 p2 pm pg pmm pmg pgg
                # test mirror plane along both diagonals at origin
                found = False
                for i in [+1, -1]:
                    spl = SymPlane(np.array([0, 0]), (abst[0]+i*abst[1]),
                                   abst)
                    if ts.is_mirror_symmetric(spl, eps):
                        found = True
                        break
                if found:
                    if toprotsym == 2:
                        planegroup = "cmm"
                    else:
                        planegroup = "cm"
                else:
                    efftype = "rectangular"
        if celltype == "rhombic" or efftype == "rhombic":
            # test mirror plane along both diagonals at origin
            found = False
            for i in [+1, -1]:
                spl = SymPlane(np.array([0, 0]), (abst[0]+i*abst[1]), abst)
                if ts.is_mirror_symmetric(spl, eps):
                    found = True
                    break
            if not found:
                efftype = "oblique"
                # since we shouldn't be here if there is no 2fold rotation,
                #   this should simply be p2...
            else:
                if toprotsym == 2:
                    planegroup = "cmm"
                else:
                    planegroup = "cm"
                    logger.warning("Unexpected point encountered in "
                                   "findSymmetry routine: FS001")
                    rp.setHaltingLevel(2)
        if celltype == "rectangular" or efftype == "rectangular":
            # test mirror plane along both unit vectors at origin
            mirs = [False, False]
            for i in range(0, 2):
                spl = SymPlane(np.array([0, 0]), abst[i], abst)
                if ts.is_mirror_symmetric(spl, eps):
                    mirs[i] = True
            if toprotsym == 2:
                if mirs[0] and mirs[1]:
                    planegroup = "pmm"
                else:
                    # test glide planes at origin
                    gldplane = None
                    for i in range(0, 2):
                        spl = SymPlane(np.array([0, 0]), abst[i], abst)
                        if ts.is_mirror_symmetric(spl, eps, glide=True):
                            spl.type = "glide"
                            gldplane = spl
                    if gldplane is not None:
                        planegroup = "pmg"
                        sl.orisymplane = gldplane
                    else:
                        # test glide plane at a/4:
                        spl = SymPlane(abst[0]/4, abst[1], abst)
                        if ts.is_mirror_symmetric(spl, eps, glide=True):
                            planegroup = "pgg"
                        else:
                            planegroup = "p2"
                if planegroup in ["pmm", "pmg", "pgg"]:
                    # each of these might be a mis-identified rcmm
                    if ts.is_rotation_symmetric((abst[0]+abst[1])/4, 2, eps):
                        planegroup = "rcmm"
            else:
                logger.warning("Unexpected point encountered in "
                               "findSymmetry routine: FS002")
                rp.setHaltingLevel(2)
        if celltype == "oblique" or efftype == "oblique":
            if toprotsym == 2:
                planegroup = "p2"
            else:
                planegroup = "p1"
                logger.warning("Unexpected point encountered in "
                               "findSymmetry routine: FS003")
                rp.setHaltingLevel(2)

    sl.planegroup = planegroup
    sl.foundplanegroup = planegroup
    if planegroup in ["pm", "pg", "cm", "rcm", "pmg"]:
        sl.foundplanegroup = planegroup+str(sl.orisymplane.par)
    else:
        sl.foundplanegroup = planegroup
    if output:
        logger.info("Found plane group: "+sl.foundplanegroup)
    if planegroup == "rcm" and not rp.SYMMETRY_FIX and not bulk:
        logger.warning(
            "The given unit cell could be reduced to half the size in a "
            "centered representation. Consider reducing the unit cell size, "
            "or using SYMMETRY_FIX to set the symmetry to p1, pm or pg.")
        rp.setHaltingLevel(1)
    if planegroup == "rcmm" and not rp.SYMMETRY_FIX and not bulk:
        logger.warning(
            "The given unit cell could be reduced to half the size in a "
            "centered representation. Consider reducing the unit cell size, "
            "or using SYMMETRY_FIX to set the symmetry to p1, p2, pm, pg, "
            "pmm, pmg, or pgg.")
        rp.setHaltingLevel(1)

    # CHECK IF USER WANTS TO MANUALLY REDUCE THE SLAB SYMMETRY, AND
    #    WHETHER THE GIVEN REDUCTION IS LEGAL
    if rp.SYMMETRY_FIX and not bulk and sl.symbaseslab is None:
        planegroup = setSymmetry(sl, rp, rp.SYMMETRY_FIX)

    return planegroup

def mirror_to_diagonal(sl, rp, abst, oriplane):
    """Rotate cell to bring oriplane along a diagonal.

    The correct direction among (11) and (1-1) is the one
    forming a 60deg angle with the symmetry plane.
    """
    directions = (abst[0]+abst[1], abst[0]-abst[1])
    angles = [angle(d, oriplane.dir) for d in directions]
    dev_from_60 =  [(a+1e-5) % np.radians(60) < 1e-3
                                    for a in angles]
    if not any(dev_from_60) or all(dev_from_60):
        err = ("The POSCAR cell could not be "
                               "rotated to a higher symmetry")
        logger.error(err)
        raise RuntimeError(err)
    idx = 0 if dev_from_60[0] else 1
    m = rotation_matrix(angles[idx], dim=3)
    sl.update_cartesian_from_fractional()
    sl.ucell_mod.append(('lmul', m))
    sl.ucell = np.dot(m, sl.ucell)
    abst = sl.ab_cell.T
    direction = (abst[0]+abst[1], abst[0]-abst[1])[idx]
    sl.collapse_cartesian_coordinates(update_origin=True)
    oriplane = SymPlane(np.array([0, 0]), direction, abst)
    logger.warning("The POSCAR unit cell was changed to a higher "
                   "symmetry form. Make sure to check beam labels for "
                   "compatibility with new unit cell.")
    rp.checklist.append("Check modified unit cell")
    rp.setHaltingLevel(2)
    return abst,oriplane


def setSymmetry(sl, rp, targetsym):
    """Sets the symmetry of the slab, based on the one found by
    findSymmetry. Can be the planegroup from findSymmetry, or a reduction
    from it."""

    def invalidDirectionMessage(rp, planegroup, targetsym, setHaltingTo=2):
        logger.warning(
            "Invalid direction given for symmetry reduction "
            "from {} to {}. Input will be ignored, proceeding without "
            "symmetry reduction.".format(planegroup, targetsym))
        rp.setHaltingLevel(setHaltingTo)

    abst = sl.ab_cell.T  # surface unit cell, transposed
    # set high symmetry
    if "[" not in sl.foundplanegroup:
        sl.planegroup = sl.foundplanegroup
    else:
        ds = re.search(r'\[(.*)\]', sl.foundplanegroup).group(1)
        d = (int(ds.split()[0]), int(ds.split()[1]))
        sl.planegroup = sl.foundplanegroup.split("[")[0]
        sl.orisymplane = SymPlane(np.array([0, 0]),
                                  d[0]*abst[0]+d[1]*abst[1], abst)
        if sl.planegroup in ["pg", "pmg"]:
            sl.orisymplane.type = "glide"
    planegroup = sl.planegroup
    if targetsym in [planegroup, "found"]:
        return planegroup
    # DICTIONARY FOR ALLOWED SYMMETRY REDUCTIONS:
    pgr = {
        'p1': [], 'p2': ['p1'], 'pm': ['p1'], 'pg': ['p1'], 'cm': ['p1'],
        'rcm': ['p1', 'pm', 'pg'], 'pmm': ['p1', 'p2', 'pm'],
        'pmg': ['p1', 'p2', 'pm', 'pg'], 'pgg': ['p1', 'p2', 'pg'],
        'cmm': ['p1', 'p2', 'cm'],
        'rcmm': ['p1', 'p2', 'pm', 'pg', 'rcm', 'pmm', 'pmg', 'pgg'],
        'p4': ['p1', 'p2'],
        'p4m': ['p1', 'p2', 'pm', 'cm', 'pmm', 'cmm', 'p4'],
        'p4g': ['p1', 'p2', 'pg', 'cm', 'pgg', 'cmm', 'p4'],
        'p3': ['p1'], 'p3m1': ['p1', 'p3', 'cm'], 'p31m': ['p1', 'p3', 'cm'],
        'p6': ['p1', 'p2', 'p3'],
        'p6m': ['p1', 'p2', 'cm', 'cmm', 'p3', 'p3m1', 'p31m', 'p6']}
    if '[' not in targetsym:
        if ((targetsym in ['pm', 'pg', 'cm', 'rcm', 'pmg']
                and planegroup != 'pmg')
                or targetsym == 'cmm' and planegroup == 'p6m'):
            logger.warning(
                "Symmetry reduction from "+planegroup+" to "+targetsym
                + " requires a direction, which was not given. Input will be "
                "ignored, proceeding without symmetry reduction.")
            rp.setHaltingLevel(1)
            targetsym = planegroup
    else:
        rgx = re.compile(r'\s*(?P<group>(pm|pg|cm|rcm|pmg|cmm))\s*\[\s*'
                         + r'(?P<i1>[-012]+)\s+(?P<i2>[-012]+)\s*\]')
        m = rgx.match(targetsym)
        targetsym = m.group('group')
        tspar = [int(m.group('i1')), int(m.group('i2'))]
        # if unit cell was changed, adapt directions
        cellchange = False
        for op in sl.ucell_mod:
            if op[0] in ['lmul', 'rmul']:
                cellchange = True
                break
        if cellchange:
            logger.warning(
                "A symmetry change was requested relative to a unit cell "
                "vector, but the unit cell has been modified. Attempting to "
                "interpret direction in the old coordinate system...")
            rp.setHaltingLevel(1)
            tspar = np.dot(np.linalg.inv(sl.ab_cell),
                           np.dot(sl.ucell_ori[:2, :2], tspar))
            for i in range(0, 2):
                tspar[i] = round(tspar[i])
    if targetsym == planegroup:
        return planegroup
    if targetsym not in pgr[planegroup]:
        logger.warning(
            "Symmetry reduction to "+targetsym+" was "
            "requested. This is not a valid symmetry reduction from "
            "the detected symmetry of "+planegroup+". Symmetry "
            + planegroup+" will be used.")
        rp.setHaltingLevel(2)
    else:
        if targetsym not in ['pm', 'pg', 'cm', 'rcm', 'pmg', 'cmm']:
            sl.planegroup = targetsym
        else:
            # NOW NEED TO GO THROUGH ALL THE MORE TRICKY TRANSFORMS
            if planegroup == 'rcm':  # reducing to: pg, pm
                if targetsym == 'pg':  # shift origin to glide plane
                    shiftv = (0.25
                              * np.dot(np.array(sl.orisymplane.par[1],
                                                -sl.orisymplane.par[0]),
                                       abst))
                    sl.translate_atoms_2d(-shiftv)
                    sl.orisymplane.type = "glide"
                    # since the origin shifts and the direction stays the
                    #  same, nothing needs to be changed about the symplane
                sl.planegroup = targetsym
            elif planegroup == 'pmm':   # reducing to: pm
                if targetsym == 'pm':
                    if (tspar[0], tspar[1]) in [(1, 0), (0, 1),
                                                (-1, 0), (0, -1)]:
                        # no harm in allowing negative directions
                        sl.planegroup = targetsym
                        sl.orisymplane = SymPlane(np.array([0, 0]),
                                                  np.dot(tspar, abst), abst)
                    else:
                        invalidDirectionMessage(rp, planegroup, targetsym)
                else:
                    sl.planegroup = targetsym
                    logger.warning("Unexpected point encountered "
                                   "in setSymmetry routine: FS004")
                    rp.setHaltingLevel(1)
            elif planegroup == 'pmg':   # reducing to: pm, pg
                if targetsym == 'pm':  # needs origin shift
                    shiftv = 0.25*np.dot(sl.orisymplane.par, abst)
                    sl.translate_atoms_2d(-shiftv)
                    sl.orisymplane = SymPlane(
                        np.array([0, 0]), np.array(sl.orisymplane.dir[1],
                                                   -sl.orisymplane.dir[0]),
                        abst)
                sl.planegroup = targetsym
            elif planegroup == 'pgg':   # reducing to: pg
                if (tspar[0], tspar[1]) in [(1, 0), (0, 1), (-1, 0), (0, -1)]:
                    shiftv = 0.25*np.dot(np.array(tspar[1], -tspar[0]), abst)
                    sl.translate_atoms_2d(-shiftv)
                    sl.orisymplane = SymPlane(np.array([0, 0]),
                                              np.dot(tspar, abst), abst)
                    sl.orisymplane.type = "glide"
                    sl.planegroup = targetsym
                else:
                    invalidDirectionMessage(rp, planegroup, targetsym)
            elif planegroup == 'cmm':   # reducing to: cm
                if (tspar[0], tspar[1]) in [(1, 1), (1, -1),
                                            (-1, -1), (-1, 1)]:
                    sl.orisymplane = SymPlane(np.array([0, 0]),
                                              np.dot(tspar, abst), abst)
                    sl.planegroup = targetsym
                else:
                    invalidDirectionMessage(rp, planegroup, targetsym)
            elif planegroup == 'rcmm':
                # reducing to: pm, pg, rcm, pmg
                if (tspar[0], tspar[1]) in [(1, 0), (0, 1), (-1, 0), (0, -1)]:
                    if targetsym in ['pm', 'pg', 'rcm']:
                        sl.orisymplane = SymPlane(np.array([0, 0]),
                                                  np.dot(tspar, abst), abst)
                        if targetsym == 'pg':
                            # shift origin to glide plane
                            shiftv = 0.25*np.dot(np.array(tspar[1],
                                                          -tspar[0]), abst)
                            sl.translate_atoms_2d(-shiftv)
                            sl.orisymplane.type = "glide"
                        sl.planegroup = targetsym
                    elif targetsym == "pmg":
                        shiftv = 0.25*(abst[0]+abst[1])
                        sl.translate_atoms_2d(-shiftv)
                        sl.orisymplane = SymPlane(np.array([0, 0]),
                                                  np.dot(tspar, abst), abst)
                        sl.orisymplane.type = "glide"
                    sl.planegroup = targetsym
                else:
                    invalidDirectionMessage(rp, planegroup, targetsym)
            elif planegroup == 'p4m':   # reducing to: pm, cm, cmm
                if targetsym == 'cmm':
                    sl.planegroup = targetsym
                elif targetsym == 'pm':
                    if (tspar[0], tspar[1]) in [(1, 0), (0, 1),
                                                (-1, 0), (0, -1)]:
                        sl.orisymplane = SymPlane(np.array([0, 0]),
                                                  np.dot(tspar, abst), abst)
                        sl.planegroup = targetsym
                    else:
                        invalidDirectionMessage(rp, planegroup, targetsym)
                elif targetsym == 'cm':
                    if (tspar[0], tspar[1]) in [(1, 1), (1, -1),
                                                (-1, -1), (-1, 1)]:
                        sl.orisymplane = SymPlane(np.array([0, 0]),
                                                  np.dot(tspar, abst), abst)
                        sl.planegroup = targetsym
                    else:
                        invalidDirectionMessage(rp, planegroup, targetsym)
            elif planegroup == 'p4g':   # reducing to: pg, cm, cmm
                allowed = True
                if targetsym == 'cmm':
                    shiftv = 0.5*abst[0]
                elif targetsym == 'pg':
                    if (tspar[0], tspar[1]) in [(1, 0), (0, 1),
                                                (-1, 0), (0, -1)]:
                        shiftv = 0.25*np.dot(np.array(tspar[1], -tspar[0]),
                                             abst)
                        sl.orisymplane = SymPlane(np.array([0, 0]),
                                                  np.dot(tspar, abst), abst)
                        sl.orisymplane.type = 'glide'
                    else:
                        allowed = False
                elif targetsym == 'cm':
                    if (tspar[0], tspar[1]) in [(1, 1), (1, -1),
                                                (-1, -1), (-1, 1)]:
                        shiftv = 0.5*abst[0]
                        sl.orisymplane = SymPlane(np.array([0, 0]),
                                                  np.dot(tspar, abst), abst)
                    else:
                        allowed = False
                if allowed:
                    sl.translate_atoms_2d(-shiftv)
                    sl.planegroup = targetsym
                else:
                    invalidDirectionMessage(rp, planegroup, targetsym)
            elif planegroup == 'p3m1':  # reducing to: cm
                if (tspar[0], tspar[1]) in [(1, -1), (-1, 1), (1, 2),
                                            (-1, -2), (2, 1), (-2, -1)]:
                    chir = 1
                    if (abs(np.dot(abst[1],
                                   np.dot(np.array([[-0.5, -np.sqrt(3)/2],
                                                    [np.sqrt(3)/2, -0.5]]),
                                          abst[0]))) > 0.01):
                        chir = -1   # left-handed unit cell -> invert rotations
                    if (tspar[0], tspar[1]) in [(1, 2), (-1, -2)]:
                        sl.rotate_unit_cell(6*chir)  # rotate 60° clockwise
                    elif (tspar[0], tspar[1]) in [(2, 1), (-2, -1)]:
                        sl.rotate_unit_cell(-6*chir)  # rotate 60° countercl.
                    abst = sl.ab_cell.T
                    sl.orisymplane = SymPlane(np.array([0, 0]),
                                              abst[0]-abst[1], abst)
                    sl.planegroup = targetsym
                else:
                    invalidDirectionMessage(rp, planegroup, targetsym)
            elif planegroup == 'p31m':  # reducing to: cm
                if (tspar[0], tspar[1]) in [(1, 0), (-1, 0), (0, 1), (0, -1),
                                            (1, 1), (-1, -1)]:
                    chir = 1
                    if abs(np.dot(abst[1], np.dot(np.array(
                            [[-0.5, -np.sqrt(3)/2], [np.sqrt(3)/2, -0.5]]),
                            abst[0]))) > 0.01:
                        chir = -1   # left-handed unit cell
                    if (tspar[0], tspar[1]) in [(1, 0), (-1, 0)]:
                        sl.rotate_unit_cell(6*chir)  # rotate 60° clockwise
                    elif (tspar[0], tspar[1]) in [(0, 1), (0, -1)]:
                        sl.rotate_unit_cell(-6*chir)  # rotate 60° countercl.
                    abst = sl.ab_cell.T
                    sl.orisymplane = SymPlane(np.array([0, 0]),
                                              abst[0]+abst[1], abst)
                    sl.planegroup = targetsym
                else:
                    invalidDirectionMessage(rp, planegroup, targetsym)
            elif planegroup == 'p6m':   # reducing to: cm, cmm
                # exclusively do 60° rotations, set mirrors accordingly
                if (tspar[0], tspar[1]) in [
                        (1, 0), (-1, 0), (0, 1), (0, -1), (1, 1), (-1, -1),
                        (1, -1), (-1, 1), (1, 2), (-1, -2), (2, 1), (-2, -1)]:
                    sl.planegroup = targetsym
                    chir = 1
                    if abs(np.dot(abst[1], np.dot(np.array(
                           [[-0.5, -np.sqrt(3)/2], [np.sqrt(3)/2, -0.5]]),
                            abst[0]))) > 0.01:
                        chir = -1   # left-handed unit cell
                    if (tspar[0], tspar[1]) in [(1, 1), (-1, -1),
                                                (1, -1), (-1, 1)]:
                        if targetsym == 'cm':
                            sl.orisymplane = SymPlane(
                                np.array([0, 0]), np.dot(tspar, abst), abst)
                    elif (tspar[0], tspar[1]) in [(1, 2), (-1, -2),
                                                  (2, 1), (-2, -1)]:
                        if (tspar[0], tspar[1]) in [(1, 2), (-1, -2)]:
                            sl.rotate_unit_cell(6*chir)  # rotate 60° clockwise
                        elif (tspar[0], tspar[1]) in [(2, 1), (-2, -1)]:
                            sl.rotate_unit_cell(-6*chir)  # rotate 60° countercl.
                        abst = sl.ab_cell.T
                        if targetsym == 'cm':
                            sl.orisymplane = SymPlane(
                                np.array([0, 0]), abst[0]-abst[1], abst)
                    elif (tspar[0], tspar[1]) in [(1, 0), (-1, 0),
                                                  (0, 1), (0, -1)]:
                        if (tspar[0], tspar[1]) in [(1, 0), (-1, 0)]:
                            sl.rotate_unit_cell(6*chir)  # rotate 60° clockwise
                        elif (tspar[0], tspar[1]) in [(0, 1), (0, -1)]:
                            sl.rotate_unit_cell(-6*chir)  # rotate 60° countercl.
                        abst = sl.ab_cell.T
                        if targetsym == 'cm':
                            sl.orisymplane = SymPlane(
                                np.array([0, 0]), abst[0]+abst[1], abst)
                else:
                    invalidDirectionMessage(rp, planegroup, targetsym)
            else:
                logger.warning("Unexpected point encountered in "
                               "findSymmetry routine: FS005")
                rp.setHaltingLevel(1)
    if targetsym == sl.planegroup:
        logger.info("The symmetry for the slab was reduced to "
                    + targetsym + ", as requested.")
    return planegroup


def enforceSymmetry(sl, rp, planegroup="fromslab",
                    movement='fromparams', rotcell=False):
    """Finds how atoms are linked to each other based on the planegroup.
    If the planegroup argument is not given, the planegroup assigned to
    the slab will be used. Otherwise, the given planegroup has to be a
    subgroup of the highest symmetry planegroup found for the slab. Sets
    movement = True or False to force or suppress recalculating atom
    positions to perfectly fit the symmetry; keep at default to follow
    SYMMETRIZE_INPUT parameter. Set rotcell = False to avoid rotating the
    unit cell such that a mirror plane is parallel to x, y, or 45°."""
    if planegroup == "fromslab":
        if sl.planegroup != "unknown":
            planegroup = sl.planegroup
        else:
            logger.warning("Call to enforceSymmetry before findSymmetry; "
                           "running findSymmetry now.")
            planegroup = sl.findSymmetry(rp)
    if movement == 'fromparams':
        nomove = not rp.SYMMETRIZE_INPUT
    else:
        if type(movement) == bool:
            nomove = not movement
        else:
            nomove = not rp.SYMMETRIZE_INPUT
            logger.warning("enforceSymmetry: Invalid 'movement' variable "
                           "passed. Using SYMMETRIZE_INPUT parameter instead.")
    eps = rp.SYMMETRY_EPS
    abst = sl.ab_cell.T  # surface unit cell, transposed

    # FIND ATOM LINKING - HERE WORK WITH sl INSTEAD OF ts, SINCE WE WANT
    #   TO ASSIGN PROPERTIES TO INDIVIDUAL ATOMS
    for at in sl:  # first put all atoms in a list of their own
        at.linklist = [at]
        at.symrefm = np.identity(2)
    if not planegroup == "p1":  # p1 has no symmetry to check for
        sl.create_sublayers(eps.z)
        sl.sort_original()
        sl.collapse_cartesian_coordinates()
        # TEST ROTATION AT ORIGIN - TESTING ONLY HIGHEST ROTATIONAL ORDER
        #    IS ENOUGH
        if planegroup not in ["p1", "pm", "pg", "cm", "rcm"]:
            if planegroup in ["p2", "pmm", "pmg", "pgg", "cmm", "rcmm"]:
                toprotsym = 2
            elif planegroup in ["p3", "p3m1", "p31m"]:
                toprotsym = 3
            elif planegroup in ["p4", "p4m", "p4g"]:
                toprotsym = 4
            elif planegroup in ["p6", "p6m"]:
                toprotsym = 6
            else:
                logger.warning("Unexpected point encountered in "
                               "enforceSymmetry routine: ES001")
                rp.setHaltingLevel(1)
            tmpslab = copy.deepcopy(sl)
            tmpslab.rotate_atoms(toprotsym)
            m = np.linalg.inv(rotation_matrix_order(toprotsym))
            for (sli, sl1) in enumerate(sl.sublayers):
                for (ati, at1) in enumerate(sl1):
                    for (atj, at2) in enumerate(tmpslab.sublayers[sli]):
                        if sl1.atlist[atj] in at1.linklist:
                            # don't check atoms that are already linked
                            continue
                        if not at1.is_same_xy(at2, eps):
                            continue
                        # combine the two linklists
                        at1.linklist.extend(sl1.atlist[atj].linklist)
                        for at3 in sl1.atlist[atj].linklist:
                            at3.symrefm = np.linalg.multi_dot([
                               np.linalg.inv(at2.symrefm),
                               at3.symrefm, m, at1.symrefm])
                            if not at3 == sl1.atlist[atj]:
                                at3.linklist = at1.linklist
                        sl1.atlist[atj].linklist = at1.linklist
                        break
        # TEST MIRROR AND GLIDE PLANES
        if planegroup not in ["p2", "p3", "p4", "p6"]:
            if planegroup in ["pm", "pg", "cm", "pmg", "rcm"]:
                # two possibilities, stored earlier
                testplane = sl.orisymplane
            elif planegroup in ["pmm", "p4m", "p6m", "rcmm"]:
                # mirror at 0
                testplane = SymPlane(np.array([0, 0]), abst[0], abst)
            elif planegroup in ["pgg", "p4g"]:
                # glide plane parallel to a at b/4
                testplane = SymPlane(abst[1]/4, abst[0], abst, ty="glide")
            elif planegroup in ["cmm", "p31m"]:
                # mirror a+b
                testplane = SymPlane(np.array([0, 0]), abst[0]+abst[1], abst)
            elif planegroup == "p3m1":
                # mirror a-b
                testplane = SymPlane(np.array([0, 0]), abst[0]-abst[1], abst)
            else:
                logger.warning("Unexpected point encountered in "
                               "enforceSymmetry routine: ES002")
                rp.setHaltingLevel(1)
            tmpslab = copy.deepcopy(sl)
            tmpslab.mirror_atoms(testplane)
            ang = angle(np.array([1, 0]), testplane.dir)
            rotm = rotation_matrix(ang)
            m = np.dot(rotm, np.dot(np.array([[1, 0], [0, -1]]),
                                    np.linalg.inv(rotm)))
            for (sli, sl1) in enumerate(sl.sublayers):
                for (ati, at1) in enumerate(sl1):
                    for (atj, at2) in enumerate(tmpslab.sublayers[sli]):
                        # don't check atoms that are already linked
                        if sl1.atlist[atj] in at1.linklist:
                            continue
                        if not at1.is_same_xy(at2, eps):
                            continue
                        # combine the two linklists
                        at1.linklist.extend(sl1.atlist[atj].linklist)
                        for at3 in sl1.atlist[atj].linklist:
                            at3.symrefm = np.linalg.multi_dot([
                                np.linalg.inv(at2.symrefm),
                                at3.symrefm, m, at1.symrefm])
                            if not at3 == sl1.atlist[atj]:
                                at3.linklist = at1.linklist
                        sl1.atlist[atj].linklist = at1.linklist
                        break
    sl.linklists = []     # re-create linklists
    for at in sl:
        if len(at.linklist) > 1 and at.linklist not in sl.linklists:
            # don't keep the linklists of length 1
            sl.linklists.append(at.linklist)
    # check that all linked atoms have the same sites
    for ll in sl.linklists:
        if len(set([at.site for at in ll])) > 1:
            logger.warning(
                "Symmetry-equivalent atoms are assigned to different sites: "
                + "; ".join(["{}: {}".format(at, at.site.label) for at in ll]))
            rp.setHaltingLevel(1)
    # FIND ALLOWED MOVE DIRECTIONS FOR ATOMS
    lockpoints = []     # full list of rotation points (no duplication)
    ori = np.array([0., 0.])
    if planegroup in ["p2", "pmm", "pmg", "pgg", "cmm", "p4", "p4m", "p4g",
                      "p6", "p6m", "rcmm"]:
        lockpoints = [ori, 0.5*abst[0], 0.5*abst[1], 0.5*(abst[0]+abst[1])]
        if planegroup in ["p6", "p6m"]:
            lockpoints.extend([(2*abst[0]+abst[1])/3,
                               (abst[0]+2*abst[1])/3])
        if planegroup == "rcmm":
            lockpoints.extend([0.25*(abst[0]+abst[1]),
                               0.75*(abst[0]+abst[1]),
                               0.25*abst[0]+0.75*abst[1],
                               0.75*abst[0]+0.25*abst[1]])
    elif planegroup in ["p3", "p3m1", "p31m"]:
        lockpoints = [ori, (2*abst[0]+abst[1])/3, (abst[0]+2*abst[1])/3]
    lockplanes = []
    # full list of mirror planes; no glide planes, since those don't restrict
    #   single atom movement
    if planegroup in ["pm", "rcm"]:
        # include duplication of planes at ori+a / ori+b to avoid having to
        #  check atom positions +- a/b
        if np.array_equal(sl.orisymplane.par, [1, 0]):
            lockplanes = [sl.orisymplane,
                          SymPlane(abst[1]/2, abst[0], abst),
                          SymPlane(abst[1], abst[0], abst, collapse=False)]
        else:
            lockplanes = [sl.orisymplane,
                          SymPlane(abst[0]/2, abst[1], abst),
                          SymPlane(abst[0], abst[1], abst, collapse=False)]
    if planegroup == "cm":
        if np.array_equal(sl.orisymplane.par, [1, 1]):
            lockplanes = [sl.orisymplane,
                          SymPlane((abst[0] - abst[1]) / 2,
                                   abst[0] + abst[1], abst, collapse=False),
                          SymPlane((abst[1] - abst[0]) / 2,
                                   abst[0] + abst[1], abst, collapse=False)]
        else:
            lockplanes = [sl.orisymplane,
                          SymPlane((abst[0] + abst[1]) / 2,
                                   abst[0] - abst[1], abst),
                          SymPlane((abst[0] + abst[1]),
                                   abst[0] - abst[1], abst, collapse=False)]
    if planegroup in ["pmm", "p4m", "rcmm"]:
        for i in range(0, 2):
            lockplanes.append(SymPlane(ori, abst[i], abst))
            lockplanes.append(SymPlane(abst[abs(i-1)]/2, abst[i], abst))
            lockplanes.append(SymPlane(abst[abs(i-1)], abst[i], abst,
                                       collapse=False))
    if planegroup in ["cmm", "p4m"]:
        lockplanes.append(SymPlane(ori, abst[0]+abst[1], abst))
        lockplanes.append(SymPlane((abst[0]+abst[1])/2, abst[0]-abst[1],
                                   abst))
        if planegroup == "cmm":
            lockplanes.extend([SymPlane(ori, abst[0]-abst[1], abst),
                               SymPlane(abst[0]+abst[1], abst[0]-abst[1],
                                        abst, collapse=False),
                               SymPlane((abst[0]-abst[1])/2, abst[0]+abst[1],
                                        abst, collapse=False),
                               SymPlane((abst[1]-abst[0])/2, abst[0]+abst[1],
                                        abst, collapse=False)])
    if planegroup == "pmg":
        if np.array_equal(sl.orisymplane.par, [1, 0]):
            lockplanes = [SymPlane(abst[0]/4, abst[1], abst),
                          SymPlane(abst[0]*3/4, abst[1], abst)]
        else:
            lockplanes = [SymPlane(abst[1]/4, abst[0], abst),
                          SymPlane(abst[1]*3/4, abst[0], abst)]
    if planegroup == "p4g":
        lockplanes = [SymPlane((abst[0]+abst[1])/4, abst[0]-abst[1], abst),
                      SymPlane((abst[0]+abst[1])*3/4, abst[0]-abst[1], abst),
                      SymPlane((abst[0]-abst[1])/4, abst[0]+abst[1], abst,
                               collapse=False),
                      SymPlane((abst[1]-abst[0])/4, abst[0]+abst[1], abst,
                               collapse=False)]
    if planegroup in ["p3m1", "p6m"]:
        lockplanes.extend([SymPlane((abst[0]+abst[1])/2, abst[0]-abst[1],
                                    abst),
                           SymPlane(ori, abst[0]+2*abst[1], abst,
                                    index2=True),
                           SymPlane(ori, 2*abst[0]+abst[1], abst,
                                    index2=True),
                           SymPlane(abst[0]+abst[1], abst[0]+2*abst[1], abst,
                                    index2=True, collapse=False),
                           SymPlane(abst[0]+abst[1], 2*abst[0]+abst[1], abst,
                                    index2=True, collapse=False)])
    if planegroup in ["p31m", "p6m"]:
        lockplanes.extend([SymPlane(ori, abst[0], abst),
                           SymPlane(ori, abst[1], abst),
                           SymPlane(abst[0], abst[1], abst, collapse=False),
                           SymPlane(abst[1], abst[0], abst, collapse=False),
                           SymPlane(ori, abst[0]+abst[1], abst)])
    ts = copy.deepcopy(sl)
    ts.project_c_to_z()
    ts.collapse_cartesian_coordinates()
    for at in ts:
        # first check points
        for p in lockpoints:
            if at.is_same_xy(p, eps):
                if not nomove:
                    at.cartpos[0:2] = p
                at.freedir = 0  # lock completely
                break
        # then if not locked yet, check planes
        if not at.freedir == 0:
            for pl in lockplanes:
                d = dist_from_line(pl.pos, pl.pos+pl.dir, at.cartpos[:2])
                if d < eps:
                    at.freedir = pl.par
                    if not nomove:  # shift atom onto plane
                        shiftv = np.array([pl.dir[1], -pl.dir[0]]) * d
                        if (dist_from_line(pl.pos, pl.pos+pl.dir,
                                           at.cartpos[:2] + shiftv)
                                > d * 1.1):
                            shiftv = -1 * shiftv
                        at.cartpos[:2] += shiftv
                    break
    for at in sl:
        at2 = [a for a in ts if a.num == at.num][0]
        at.freedir = at2.freedir
        at.cartpos = at2.cartpos
    # average positions for linked atoms
    if not nomove and not planegroup == "p1":
        sl.collapse_cartesian_coordinates()
        releps = [eps / np.linalg.norm(abst[j]) for j in range(0, 2)]
        mvslabs = []
        if planegroup not in ["pm", "pg", "cm", "rcm"]:
            for i in range(0, toprotsym-1):
                if len(mvslabs) == 0:
                    tmpslab = copy.deepcopy(sl)
                else:
                    tmpslab = copy.deepcopy(mvslabs[-1])
                tmpslab.rotate_atoms(toprotsym, ori)
                mvslabs.append(tmpslab)
        if planegroup not in ["p2", "p3", "p4", "p6"]:
            tmpslab = copy.deepcopy(sl)
            tmpslab.mirror_atoms(testplane)
            mvslabs.append(tmpslab)
            if planegroup not in ["pm", "pg", "cm", "rcm"]:
                for i in range(0, toprotsym-1):
                    tmpslab = copy.deepcopy(mvslabs[-1])
                    tmpslab.rotate_atoms(toprotsym, ori)
                    mvslabs.append(tmpslab)
        for (llind, ll) in enumerate(sl.linklists):
            for at in ll:
                psum = np.copy(at.cartpos)
                pn = 1
                for ms in mvslabs:
                    found = False
                    for atind in range(0, len(ll)):
                        at2 = ms.linklists[llind][atind]
                        complist = [at2.cartpos[:2]]
                        for j in range(0, 2):
                            if abs(at2.pos[j]) < releps[j]:
                                complist.append(at2.cartpos[:2]+abst[j])
                            if abs(at2.pos[j]-1) < releps[j]:
                                complist.append(at2.cartpos[:2]-abst[j])
                        if len(complist) == 3:
                            # corner - add the diagonally opposed one
                            complist.append(complist[1]+complist[2]
                                            - complist[0])
                        for p in complist:
                            if np.linalg.norm(p - at.cartpos[:2]) < eps:
                                psum += np.append(p, at2.cartpos[2])
                                pn += 1
                                found = True
                                break
                        if found:
                            break
                if pn != len(mvslabs) + 1:
                    logger.warning(
                        "During symmetrization of the slab, not all "
                        "symmetry-equivalent atoms were found for {}. Atom "
                        "will not be symmetrized.")
                else:
                    at.cartpos = psum / pn
    sl.collapse_cartesian_coordinates(update_origin=True)
    if not rotcell:
        # !!! THIS IS NOW THE DEFAULT, rotcell is NEVER true.
        #  if it stays like this, consider deleting the following lines..
        return
    # after everything else is done, rotate unit cell (in x,y without
    #   changing fractional coordinates) if necessary:
    # because TensErLEED is faster if mirror planes are along x, y, or x+y
    mirrordirs = []
    # if a or b are already along one of the "allowed" directions, we don't
    #   need to do anything
    if planegroup in ["cm", "cmm", "p3m1", "p31m", "p6", "p6m"]:
        # rotate to one of the diagonals
        mirrordirs.append((abst[0]-abst[1])
                          / np.linalg.norm(abst[0]-abst[1]))
        mirrordirs.append((abst[0]+abst[1])
                          / np.linalg.norm(abst[0]+abst[1]))
    elif planegroup in ["pm", "rcm", "pmm", "pmg", "rcmm", "p4m", "p4g"]:
        mirrordirs.append(abst[0]/np.linalg.norm(abst[0]))
        mirrordirs.append(abst[1]/np.linalg.norm(abst[1]))
        if planegroup in ["p4m", "p4g"]:
            mirrordirs.append((abst[0]+abst[1])
                              / np.linalg.norm(abst[0]+abst[1]))
            mirrordirs.append((abst[0]-abst[1])
                              / np.linalg.norm(abst[0]-abst[1]))
    if len(mirrordirs) > 0:
        found = False
        for d in mirrordirs:
            for i in range(0, 2):
                if abs(np.dot(np.array([1, 0]), d)-1) < 0.001:
                    found = True
        if not found:
            ang = angle(np.array([1, 0]), mirrordirs[0])
            rotm = rotation_matrix(ang, dim=3)
            sl.ucell = np.dot(rotm, sl.ucell)
            for i in range(0, 3):
                for j in range(0, 3):
                    if abs(sl.ucell[i, j]) < 1e-6:
                        sl.ucell[i, j] = 0
            sl.update_cartesian_from_fractional()
            # modify BEAM_INCIDENCE
            if rp.THETA != 0:
                logger.debug("Modifying BEAM_INCIDENCE parameter")
                rp.PHI += np.degrees(ang)
                parameters.modify(rp, "BEAM_INCIDENCE")
    return


def getSymBaseSymmetry(sl, rp):
    """Runs the symmetry search for the symbaseslab, then transfers atom
    linking to translationally equivalent atoms in the extended slab."""
    if sl.symbaseslab is None:
        logger.error("getSymBaseSymmetry: No symmetry base slab defined.")
        raise RuntimeError("getSymBaseSymmetry called without symmetry base "
                           "slab.")
    if sl.symbaseslab.planegroup == "unknown":
        findSymmetry(sl.symbaseslab, rp, forceFindOri=True)
        enforceSymmetry(sl.symbaseslab, rp, rotcell=False)
    sl.update_cartesian_from_fractional()
    for ll in sl.symbaseslab.linklists:
        newll = []
        for ssl_at in ll:
            at = [a for a in sl if a.num == ssl_at.num][0]
            newll.append(at)
            at.linklist = newll
            at.symrefm = np.copy(ssl_at.symrefm)
    for at in [at for at in sl if at.duplicate_of]:
        at.duplicate_of.linklist.append(at)
        at.linklist = at.duplicate_of.linklist
        at.symrefm = np.copy(at.duplicate_of.symrefm)
        if rp.SYMMETRIZE_INPUT:
            cv = at.cartpos[:2] - at.duplicate_of.cartpos[:2]
            v = np.append(np.round(np.dot(np.linalg.inv(
                                     sl.symbaseslab.ab_cell), cv)), 0.)
            at.cartpos = (at.duplicate_of.cartpos
                          + np.dot(sl.symbaseslab.ucell, v))
    sl.update_fractional_from_cartesian()
    sl.linklists = []
    for at in sl:
        if len(at.linklist) > 1 and at.linklist not in sl.linklists:
            sl.linklists.append(at.linklist)



"""
======================================
  ViPErLEED Graphical User Interface
======================================
 *** module guilib.leedsim.classes.beams ***

Defines the LEEDSymetryDomains class, used for constructing a LEED pattern

Author: Michele Riva
Created: 2021-03-13
"""

import copy
from warnings import warn as warning   # eventually will replace with logging

import numpy as np

from viperleed import guilib as gl


class LEEDEquivalentBeams:                                                      ## TODO: fix docstring
    """
    This class is just a processor for finding beams that are equivalent among
    the ones used for instantiation. The beams originate from the same bulk,
    whose reciprocal-space basis is also passed during instantiation.

    The class is instantiated with one or more beam dictionaries lists. Each of
    the beam dictionaries in the list represents a 'building block', e.g., the
    domains originating from symmetry or different structural domains.
    Each of the dictionaries should have the following structure:
    keys : tuple
        each key is a beam index (with respect to the basis passed) of a LEED
        beam
    values : set of tuples
        list of beams equivalent to key
    In addition a list of glide-extinct beams can also be passed upon
    instantiation. If passed, it should have the same length as the domains
    dictionaries.
    
    Alternatively, also list of LEEDEquivalentBeams instances can be passed in
    the constructor, in which case the information will be gathered from them,
    and there is no need to pass the bulk_basis or the lists of extinct beams
    
    The primary goal of the class is to combine all the dictionaries given as
    input into new data structures that contain the beam equivalences when all
    the domains given upon instantiation are combined in an incoherent fashion
    
    Since the calculations are relatively slow, and the combinations are many,
    it's a good idea to cache instances with a hash that is computed such
    that it is unique for the geometry considered. This gives speedups of more
    than a factor 100 upon re-using an already calculated geometry
    
    This class is intended only for internal use. TRUE???
    """

    hash_keys = ('basis', 'superlattice', 'domain_ids', 'angle_key')

    __cache = {}
    def __init__(self, domains, **kwargs):
        # Do some checking of the parameters passed
        # see if the domains passed are all instances of LEEDEquivalentBeams
        if all(isinstance(dom, LEEDEquivalentBeams) for dom in domains):
            # check that the elements passed are consistent with one another,
            # i.e. that the lattice bases are compatible
            if not all(np.allclose(dom.basis, domains[0].basis)
                       for dom in domains):
                raise ValueError("LEEDEquivalentBeams: domains have different"
                                 "bulk bases, and cannot be combined")
            # In this case, treat the current instance as a superposition
            # of the inputs. This allows to treat different STRUCTURES, each
            # on the same bulk, and each possibly containing several SYMMETRY-
            # related domains
            basis = domains[0].basis
            extinct_lists = [dom.extinct for dom in domains]
            domains = [dom.equivalent_beams_dict for dom in domains]

        elif not all(isinstance(dom, dict) for dom in domains):
            # The only other option is passing a list of beam-equivalence
            # dictionaries, as well as all the (now mandatory) parameter basis
            raise TypeError("LEEDEquivalentBeams: only list of dictionaries or "
                            "list of LEEDEquivalentBeams are acceptable. Not "
                            "possible to use mixed lists")
        else:
            try:
                basis = kwargs['basis']
            except KeyError:
                raise TypeError("LEEDEquivalentBeams: missing mandatory basis"
                                "parameter")
            extinct_lists = kwargs.pop('extinct_lists', []*len(domains))

        if np.shape(basis) != (2, 2):
            raise ValueError("LEEDEquivalentBeams: invalid bulk_basis "
                             + f"{basis}. ".replace('\n', '')
                             + "Expected a 2x2 array-like")
        if len(extinct_lists) != len(domains):
            raise ValueError("LEEDEquivalentBeams: incompatible number of "
                             + f"domains ({len(extinct_lists)}) found in "
                             + f"extinct_lists. Expected {len(domains)}")
        domain_ids = kwargs.get('domain_ids', range(len(domains)))
        
        self.__hash_dict = {}      # set by self.__update_hash
        
        # re-prepare the kwargs to update the hash dictionary, updating values
        # that may have changed during this initialization
        kwargs['basis'] = basis
        kwargs['domain_ids'] = domain_ids
        self.__update_hash(**kwargs)

        if self.is_cached:
            # skip all the calculations, and take the attributes from the cached
            # object
            cached = self.__cache[self.hash]
            self.__basis = cached.basis
            self.__eq_beams_dict = cached.equivalent_beams_dict
            self.__extinct = cached.extinct
            self.__indexed_beams = cached.indexed_beams
            return None

        # if never calculated before, or hash cannot be determined due to lack
        # of input, calculate again. If it can be hashed, cache self
        self.__basis = basis
        self.__eq_beams_dict = {}  # set in __build_beam_equivalence_dict
        self.__extinct = []        # set in __index_beams
        beam_groups = self.__build_beam_equivalence_dict(domains)
        beam_groups = self.__sort_beams(beam_groups)
        self.__indexed_beams = self.__index_beams(beam_groups, domains,
                                                  extinct_lists, domain_ids)
        if self.hash:
            self.__cache[self.hash] = self

    @staticmethod
    def clear_cache():
        """
        Completely clears the hash table of LEEDEquivalentBeams instances.
        This may be useful in case a completely new bulk structure is loaded,
        in case memory usage starts being an issue.
        """
        LEEDEquivalentBeams.__cache = {}

    @property
    def basis(self):
        """
        Basis vectors as a 2x2 iterable. a = basis[0], b = basis[1]
        """
        return self.__basis

    @property
    def extinct(self):
        """
        List of truly extinct beams, each is a tuple. Beams are truly extinct
        if they do not overlap, or if they originate from the superposition
        of extinct beams
        """
        return self.__extinct

    @property
    def equivalent_beams_dict(self):
        """
        Returns
        -------
        dict
            keys : tuple
                beam i
            values : set of tuples
                beams equivalent to beam i, taking the superposition of domains
                into account
        """
        return self.__eq_beams_dict

    @property
    def indexed_beams(self):
        return self.__indexed_beams

    @property
    def hash(self):
        """
        Computes the hash of the geometry defined during instantiation
        """
        if not self.__hash_dict:
            return 0
        return hash(tuple(self.__hash_dict.values()))

    @property
    def is_cached(self):
        return self.hash and (self.hash in self.__cache)

    def __update_hash(self, **kwargs):
        missing = [key for key in self.hash_keys if key not in kwargs]
        if missing:
            warning(f"LEEDEquivalentBeams: missing {missing}. No hash possible")
            return None
        self.__hash_dict = kwargs
        for k, v in self.__hash_dict.items():
            if k in ('basis', 'superlattice'):
                v = zip(*v)
            if k in ('basis', 'superlattice', 'domain_ids'):
                self.__hash_dict[k] = tuple(v)

    def __build_beam_equivalence_dict(self, beams_by_domain):
        """
        Parameters
        ----------
        beams_by_domain : list of dict
            Each dict is of the form
            key : tuple
                beam j of domain i
            value : set of tuples
                beams equivalent to beam j, possibly only considering beams
                of the domain, and certainly without accounting for
                superposition between domains

        Returns
        -------
        list of sets of tuples
            each of the elements of the list, is the collection of all beams
            that are equivalent to one another as a result of the superposition
            of the domains
            NB: this method also sets the instance attribute __eq_beams_dict,
                i.e., a dictionary of {beam: {beams equivalent to beam}}
        """
        # store a deepcopy of the dictionary, as later on we will pop some
        # of the elements, and we don't want to screw with the input
        beams_by_domain_copy = copy.deepcopy(beams_by_domain)
    
        # Flatten the list of all beams, keeping only uniques
        flat_beams = set(beam for beams in beams_by_domain for beam in beams)
        
        # Now figure out which of the beams are symmetry equivalent
        # taking superposition of spots into account:
        # spots originating from beams that are equivalent within each domain,
        # are equivalent only if all domains contribute with equivalent beams 
        # in all spots
        #
        # For example (symmetry domains on hexagonal bulk):
        #   dom1: (1 0), (-1 0), (1 -1), (-1 1) equivalent
        #   dom2: (1 0), (-1 0), (0 -1), (0 1)  equivalent
        #   dom3: (1 0), (-1 0)                 equivalent
        #   --> (1 0) and (-1 0) are equivalent
        #
        #   dom1: (1 0), (-1 0), (1 -1), (-1 1)  equivalent
        #   dom2:                (1 -1), (-1 1)  equivalent
        #   dom3: (0 1), (0 -1), (1 -1), (-1 1)  equivalent
        #   --> (1 -1) and (-1 1) are equivalent, but are
        #       not equivalent to [(1 0), (-1 0)]
        #
        # Another example (structural domains on square bulk):
        # p(2x2)-pmm + c(2x2)-pm[1 0] -- take spot (1 1)
        #   
        #   p(2x2): (1 1), (1 -1), (-1  1), (-1 -1)  equivalent
        #   c(2x2): (1 1), (1 -1)                    equivalent
        #   --> (1 1) and (1 -1) are equivalent
        #
        #   p(2x2): (1 1), (1 -1), (-1 1), (-1 -1)   equivalent
        #   c(2x2):                (-1 1), (-1 -1)   equivalent
        #   --> (-1 1) and (-1 -1) are equivalent, but are
        #       not equivalent to [(1 1), (1 -1)]
        beam_equivalence = []
        for beam in flat_beams:
            # from each structure, if there is a beam <beam>, take the list of all
            # those equivalent to it.
            # Example: beam = (1 1)
            #   -> take [[(1 1), (1 -1), (-1 -1), (1 -1)],  from p(2x2)
            #            [(1 1), (1 -1)]]                   from c(2x2)
            beam_sets = [beams_dict[beam]
                         for beams_dict in beams_by_domain_copy
                         if beam in beams_dict]
            if beam_sets:
                # if there are beams to process in the list, take the set
                # intersection, i.e., find all those in common to all structures
                common_beams = set.intersection(*beam_sets)
                
                # now "mark as processed" in all the structures the beams coming
                # from the intersection by removing them, and set the correct
                # keys of beam_equivalence
                for processed_beam in common_beams:
                    for unprocessed_beams in beams_by_domain_copy:
                        unprocessed_beams.pop(processed_beam, None)
                    if processed_beam in self.__eq_beams_dict:  # TODO: remove this after it's clear that it does not happen
                        raise KeyError("TEMP CHECK: setting multiple times the same key!")
                    self.__eq_beams_dict[processed_beam] = common_beams
                beam_equivalence.append(common_beams)
        return beam_equivalence

    def __sort_beams(self, beam_groups):
        """
        Sort beams given as a list of iterables. Each of the elements in the
        list is treated as a group of equivalent beams
        """
        # sort within each equivalence group and by energy
        return sorted(
            (sorted(beams, key=self.__sort_hk) for beams in beam_groups),
            key=self.__sort_energy
        )
    
    def __index_beams(self, beam_groups, beams_by_domain, extinct, domain_ids):
        """
        Given the (sorted) list of beam groups, assign to each of the
        groups one index. Also, determine which of the original domains
        contribute to the overlapped spots, and which of these contributions
        come from extinct beams. The domains (both in the overlapping and in
        the extinct lists) are labeled according to the domain_ids given
        NB: this method also sets up the __extinct instance attribute, that
            contains all the beams that are truly extinct, as a result of the
            superposition of extinct beams

        Parameters
        ----------
        beam_groups : iterable of iterables of tuples
            List of beam groups, already sorted
        beams_by_domain : list of iterables of tuples
            each entry is a domain, and contains all beams of the domain
        extinct : list of tuples
            Each entry in the list contains the beams that are extinct in each
            domain
        domain_ids : iterable
            contains names for the domains. Typically would be integers
        """
        
        indexed_beams = []
        for i, beams in enumerate(beam_groups):
            for beam in beams:
                # find which domains overlap
                overlapping = [d
                               for d, domain_beams in enumerate(beams_by_domain)
                               if beam in domain_beams]
                group_idx = i
                
                # figure out whether the beam is extinct in all the structures
                # that overlap, in which case the index goes negative
                extinct_domains = [d for d in overlapping if beam in extinct[d]]
                n_extinct = len(extinct_domains)
                if n_extinct == len(overlapping) and n_extinct > 0:
                    group_idx *= -1
                    self.__extinct.append(beam)
                indexed_beams.append((beam, group_idx,
                                      [domain_ids[d] for d in overlapping],
                                      [domain_ids[d] for d in extinct_domains]))
        # TODO: This format is not great. Since I'm always using it unzipped, it
        # may be worth just returning it unzipped in the first place, maybe as
        # a dictionary just to have some reasonable names, like 'beams',
        # 'groups', 'overlap domains', 'extinct domains'
        return indexed_beams

    def __sort_hk(self, index):
        sortH, sortK = index
        return -sortH-sortK, -sortH
    
    def __sort_energy(self, indices):
        # assumes that the indices passed represent beams that are equivalent to
        # one another, i.e., they all have the same length
        g = np.dot(indices[0], self.basis)
        # sort by energy, i.e., by ||g||^2:
        return g[0]**2 + g[1]**2






"""
======================================
  ViPErLEED Graphical User Interface
======================================
 *** module guilib.leedsim.classes.realspace ***

Defines the RealSpace class, used for displaying the real-space lattice

Author: Michele Riva
Created: 2021-03-13
"""

import numpy as np

from viperleed import guilib as gl

class RealSpace():
    def __init__(self, params):
        params = gl.LEEDParameters(params)

        self.superlattice = params['SUPERLATTICE']

        # Set up the lattices
        surf_basis = params['surfBasis']

        # self.fov is the real space field of view.
        # contains at least 4 unit cells
        self.fov = 4.2*max(surf_basis.ravel())
        self.surf = gl.Lattice(surf_basis,
                               space='real',
                               group=params['surfGroup'],
                               limit=self.fov)
        self.bulk = gl.Lattice(self.bulk_basis,
                               space='real',
                               group=params['bulkGroup'],
                               limit=self.fov)
        self.bulk.group.screws_glides = (params['bulk3Dsym'],
                                         self.bulk.cell_shape)

    @property
    def bulk_basis(self):
        """
        Real-space basis of the bulk lattice
        """
        if hasattr(self, 'bulk'):
            return self.bulk.basis
        return np.dot(np.linalg.inv(self.superlattice),
                      self.surf.basis).round(10)

    def angle_for_horizontal_bulk(self, direction):
        if direction not in (0, 1):
            raise ValueError("First positional argument of "
                             "angle_for_horizontal_bulk() "
                             f"should be 0 or 1. Got {direction!r} instead.")

        basis = self.bulk.basis[direction]
        # NB: np.arctan2 takes first the 'y' then the 'x'!
        theta = np.arctan2(basis[1], basis[0])

        return np.degrees(theta)




"""Version class to keep track of version numbers internally.

Taken originally from guilib."""

__authors__ = (
    'Michele Riva (@michele-riva)',
    'Alexander M. Imre (@amimre)',
    )
__copyright__ = 'Copyright (c) 2019-2024 ViPErLEED developers'
__created__ = '2024-04-15'
__license__ = 'GPLv3+'

import re

_DOTS_OR_DIGITS = re.compile(r"[.0-9]+")

class Version:
    """Simple class to easily handle versions."""

    def __init__(self, major, minor=-1, patch=-1):
        """Initialize instance.

        Parameters
        ----------
        major : int, float, str, or Version
            When an integer, interpret as the "major" part of
            version; when a string, it should be of the form
            "<major>.<minor>.<patch>", with minor and patch
            optional, and each field be int-able; when a float,
            interpret the whole as the major and the decimal
            as the minor.
        minor : int, optional
            Minor portion of the version. This is discarded if major
            is given as a string, a float, or a Version. If negative
            or not given, it is interpreted as "zero".
        minor : int, optional
            Patch portion of the version. This is discarded if major
            is given as a string, a float, or a Version. If negative
            or not given, it is interpreted as "zero".

        Raises
        ------
        TypeError
            If any of the inputs is not an integer. If major is given
            as a string, the exception is raised if one of the parts
            is not int-able.
        ValueError
            If major is an empty string, if it does not conform to
            the <number>[.<number>[.<number>]] syntax, if <major>
            is negative, if <minor> is not given but <patch> is.
        """
        _name = self.__class__.__name__
        if isinstance(major, float):
            major = str(major)
        if isinstance(major, str):
            major, minor, patch = self.__from_string(major)
        elif isinstance(major, Version):
            major, minor, patch = major

        if not all(isinstance(p, int) for p in (major, minor, patch)):
            raise TypeError(f"{_name}: Invalid type for one of the parts")
        if major < 0:
            raise ValueError(f"{_name}: version cannot be negative")

        # pylint: disable=chained-comparison
        # pylint would probably like the next one to be
        # minor < 0 < patch, but it feels less readable
        if patch > 0 and minor < 0:
            raise ValueError(f"{_name}: Cannot have 'patch' without 'minor'")
        self.__has_parts = (True, minor >= 0, patch >= 0)
        self._parts = (major, max(minor, 0), max(patch, 0))

    @property
    def major(self):
        """Return the major part of self."""
        return self._parts[0]

    @property
    def minor(self):
        """Return the minor part of self (if present) or RuntimeError."""
        if not self.__has_parts[1]:
            raise RuntimeError(f"{self.__class__.__name__} has no minor.")
        return self._parts[1]

    @property
    def patch(self):
        """Return the minor part of self (if present) or RuntimeError."""
        if not self.__has_parts[2]:
            raise RuntimeError(f"{self.__class__.__name__} has no patch.")
        return self._parts[2]

    def __eq__(self, other):
        """Return True if self == other, NotImplemented otherwise."""
        if not isinstance(other, self.__class__):
            try:
                other = self.__class__(other)
            except TypeError:
                return NotImplemented
        if self._parts == other._parts:
            return True
        return NotImplemented

    def __iter__(self):
        """Return an iterable version of self."""
        return iter(self._parts)

    def __lt__(self, other):
        """Return whether self < other."""
        if not isinstance(other, self.__class__):
            try:
                other = self.__class__(other)
            except TypeError:
                return NotImplemented
        return self._parts < other._parts

    def __le__(self, other):
        """Return whether self <= other."""
        if not isinstance(other, self.__class__):
            try:
                other = self.__class__(other)
            except TypeError:
                return NotImplemented
        return self._parts <= other._parts

    def __gt__(self, other):
        """Return whether self > other."""
        if not isinstance(other, self.__class__):
            try:
                other = self.__class__(other)
            except TypeError:
                return NotImplemented
        return self._parts > other._parts

    def __ge__(self, other):
        """Return whether self > other."""
        if not isinstance(other, self.__class__):
            try:
                other = self.__class__(other)
            except TypeError:
                return NotImplemented
        return self._parts >= other._parts

    def __str__(self):
        """Return a string representation of self."""
        return ".".join(str(p)
                        for p, had_part in zip(self._parts, self.__has_parts)
                        if had_part)

    def __repr__(self):
        """Return a representation of self."""
        return f"{self.__class__.__name__}({self})"

    def __from_string(self, _as_string):
        """Return major, minor, patch from a string."""
        _name = self.__class__.__name__
        if not _as_string:
            # Empty version
            raise ValueError(f"{_name}: version cannot be an empty string")
        if not _DOTS_OR_DIGITS.match(_as_string):
            raise ValueError(
                f"{_name}: version can contain only digits and dots"
                )
        major, *rest = (int(v) for v in _as_string.split('.'))
        try:
            minor = rest.pop(0)
        except IndexError:
            minor = -1
        try:
            patch = rest.pop(0)
        except IndexError:
            patch = -1
        return major, minor, patch



"""
===============================================
      ViPErLEED Graphical User Interface
===============================================
 *** module guilib.leedsim.dialogbulk3dsym ***

Created: 2021-05-17
Author: Michele Riva

"""

import numpy as np

import PyQt5.QtCore as qtc
import PyQt5.QtGui as qtg
import PyQt5.QtWidgets as qtw

from viperleed import guilib as gl


TEXT_FOR_OP = {gl.PlaneGroup.C6: '6-fold screw/rotations',
               gl.PlaneGroup.C4: '4-fold screw/rotations',
               gl.PlaneGroup.C3: '3-fold screw/rotations',
               gl.PlaneGroup.C2: '2-fold screw/rotation',
               gl.PlaneGroup.Mx: 'mirror/glide across [1, 0]',
               gl.PlaneGroup.My: 'mirror/glide across [0, 1]',
               gl.PlaneGroup.M11: 'mirror/glide across [1, 1]',    # == M45
               gl.PlaneGroup.M1m1: 'mirror/glide across [1, -1]',  # == Mm45
               gl.PlaneGroup.M10: 'mirror/glide across [1, 0]',
               gl.PlaneGroup.M01: 'mirror/glide across [0, 1]',
               gl.PlaneGroup.M21: 'mirror/glide across [2, 1]',
               gl.PlaneGroup.M12: 'mirror/glide across [1, 2]'}


class Bulk3DSymDialog(qtw.QDialog):
    """Dialog to input extra symmetry operations for bulk.


    Attributes
    ----------
    None.

    Methods
    -------
    update_operations(bulk)
        Update widgets to include all admissible operations
        that may be selected by the user. Should be called
        before opening/showing/executing the dialog.
    """

    def __init__(self, parent=None):
        super().__init__(parent)

        self.setWindowModality(qtc.Qt.WindowModal)
        flags = self.windowFlags()
        flags &= ~qtc.Qt.WindowCloseButtonHint #disable close button
        self.setWindowFlags(flags)
        self.setWindowTitle('Add bulk symmetry operations')

        # sub-widgets of which we need to keep a reference
        self.__all_widgets = {'conserve_sym': qtw.QCheckBox(),
                              'extra_ops': qtw.QListView(),
                              'buttons': {'done': qtw.QPushButton(),
                                          'cancel': qtw.QPushButton()}}

        # The following are set update_operations():
        self.__bulk = None      # gl.Lattice
        self.__extra_ops = []   # extra operations that may be selected

        self.__compose()
        self.__connect()

    def __compose(self):
        """Place widgets in the right spots."""
        description = qtw.QLabel()
        description.setText("Pick extra symmetry operations for the bulk. "
                            "\nThese are useful to account for symmetries "
                            "that are not captured by the plane group, e.g., "
                            "screw axes perpendicular to the surface, and "
                            "glide operations where mirroring is across a "
                            "plane perpendicular to the surface, and "
                            "translation has a pure z component.\nThey are "
                            "useful in case the sample has steps, and the "
                            "reconstruction rotates (or is mirrored) from one "
                            "terrace to the next [e.g., Fe3O4(001) or Si(001)]."
                            "\n")
        description.setWordWrap(True)
        description.setFont(gl.AllGUIFonts().labelFont)

        widgets = [description]

        sym_conserve = self.__all_widgets['conserve_sym']
        sym_conserve.setText("Constrain symmetry to bulk cell")
        sym_conserve.setFont(gl.AllGUIFonts().labelFont)
        sym_conserve.setChecked(True)
        widgets.append(sym_conserve)

        # Until we figure out how to handle the input (see Issue #15)
        # leave it checked and disabled
        sym_conserve.setEnabled(False)

        # Actual list of available extra symmetry operations
        widgets.append(self.__all_widgets['extra_ops'])

        # Done and Cancel buttons
        buttons = self.__all_widgets['buttons']
        buttons['done'].setText('&Done')
        buttons['cancel'].setText('&Cancel')
        widgets.extend([but for but in buttons.values()])

        # Make sure every graphical aspect of widgets is up to date
        for widget in widgets:
            widget.ensurePolished()

        # Change resizing policy for some widgets:
        # - Done and Cancel buttons should not resize
        for but in buttons.values():
            but.setSizePolicy(qtw.QSizePolicy.Fixed, qtw.QSizePolicy.Fixed)

        # Place widgets into the layout
        layout = qtw.QVBoxLayout()
        for widget in widgets:
            if widget in buttons.values():
                continue
            layout.addWidget(widget)

        buttons_layout = qtw.QHBoxLayout()
        for button in buttons.values():
            buttons_layout.addWidget(button)
            buttons_layout.setAlignment(qtc.Qt.AlignRight)
        layout.addLayout(buttons_layout)

        self.setLayout(layout)

    def __connect(self):
        """Hook up signals."""
        self.__all_widgets['extra_ops'].clicked.connect(self.__item_clicked)

        buttons = self.__all_widgets['buttons']
        buttons['cancel'].clicked.connect(self.close)
        buttons['done'].clicked.connect(self.__done_pressed)

    def __get_symmetry_related_operations(self, operation):
        """Return operations symmetry-related to the given one.

        'Related' means it is an operation that results from
        applying (operation*another_operation) where
        another_operation is one of those selected, or one
        of the 2D plane group operations of the bulk. The list
        will not contain operation itself.

        Parameters
        ----------
        operation : tuple
            2x2 matrix of a symmetry operation.

        Returns
        -------
        list
        """
        # This is very hard to do properly, when allowing the user
        # to input arbitrary operations as one would have to check
        # all the combinations of operations.
        # TODO: do this later when we figure out how to handle this.

        #       For now, look through the gl.PlaneGroup(s) for the
        #       current bulk.cell_shape, and pick the smallest group
        #       that contains all the selected operations (+ those of
        #       the plane group itself). Then use the operations of
        #       this group to automatically select the ones that are
        #       symmetry-related.
        extras = self.__all_widgets['extra_ops'].model().item
        all_ops = (*self.__bulk.group.operations(),
                   *[op
                     for i, op in enumerate(self.__extra_ops)
                     if extras(i).checkState()])

        # Get the smallest group that contains all the operations
        for group in gl.PlaneGroup.groupsForShape[self.__bulk.cell_shape]:
            group_ops = gl.PlaneGroup(group).operations()
            if all(op in group_ops for op in all_ops):
                break
            group_ops = []

        # For now stick to the easy ones: C6 --> C3 & C2; C4 -> C2
        related = []
        for related_op in group_ops:
            if related_op == operation:
                continue  # skip the one passed
            if related_op not in self.__extra_ops:
                continue  # skip those not in list
            related.append(related_op)

        return related

        # if operation in (gl.PlaneGroup.C6, gl.PlaneGroup.C4):
            # for related_op in (gl.PlaneGroup.C2, gl.PlaneGroup.C3):
                # try:
                    # related.append(self.__extra_ops.index(related_op))
                # except ValueError:
                    # pass

    def __done_pressed(self):
        """Return with code == qtw.QDialog.Accepted."""
        super().done(qtw.QDialog.Accepted)
        # self.accept()
        # self.close()

    def __item_clicked(self, index):
        """Toggle check-box of clicked item and those related.

        This is a slot to be connected to QListView.clicked.

        Parameters
        ----------
        index : QtCore.QModelIndex
        """
        item = self.__all_widgets['extra_ops'].model().itemFromIndex(index)

        if not item.isEnabled():
            return

        # toggle check state
        if item.checkState() == qtc.Qt.Unchecked:
            checked = qtc.Qt.Checked
        else:
            checked = qtc.Qt.Unchecked
        item.setCheckState(checked)

        # If the user checks an operation, others related to it via
        # either the bulk operations or others that are already selected
        # must also get selected automatically (and disabled).
        related_ops = self.__get_symmetry_related_operations(
            self.__extra_ops[index.row()]
            )

        items = self.__all_widgets['extra_ops'].model().item
        for op in related_ops:
            idx = self.__extra_ops.index(op)
            items(idx).setCheckState(checked)
            items(idx).setEnabled(not bool(checked))

    @property
    def n_extra_ops(self):
        """Return the number of extra operations that may be selected."""
        model = self.__all_widgets['extra_ops'].model()
        if model.item(0):
            return model.rowCount()
        return 0

    def update_operations(self, bulk):
        """Place admissible extra operations in the extra_ops widget.

        This method should be called before .show(), .open(), or
        .exec() to have the list of symmetry operations up to date.

        Parameters
        ----------
        bulk : gl.Lattice

        Returns
        -------
        None.
        """
        self.__bulk = bulk
        bulk_ops = self.__bulk.group.operations()

        # Get the plane group with most operations given the
        # current cell shape (always the last in groupsForShape)
        largest_group = gl.PlaneGroup(
            bulk.group.groupsForShape[bulk.cell_shape][-1]
            )
        self.__extra_ops = []
        for op in largest_group.operations():
            # Skip those that are already in the bulk group
            if op in bulk_ops:
                continue
            # Skip those that are the inverse of one that is
            # already present. They will always be added, but
            # we don't want the user to bother selecting both.
            # They are: Cm6 (for C6), Cm3 (for C3) and Cm4 (for C4)
            if op in (gl.PlaneGroup.Cm6, gl.PlaneGroup.Cm3, gl.PlaneGroup.Cm4):
                continue
            self.__extra_ops.append(op)

        # Add also those screws_glides that we may not have
        for op in bulk.group.screws_glides:
            if op in self.__extra_ops:
                continue
            self.__extra_ops.append(op)

        n_extra = len(self.__extra_ops)

        model = qtg.QStandardItemModel(n_extra, 1)  # rows, col
        for r, extra_op in enumerate(self.__extra_ops):
            txt = TEXT_FOR_OP.get(extra_op, None)

            # Handle arbitrary input
            if txt is None:
                # Not one of the known operations.
                # For now place the matrix straight as is.
                # TODO: adapt once we figure out how to handle this
                txt = str(extra_op)

            item = qtg.QStandardItem(txt)
            item.setFlags(qtc.Qt.ItemIsEnabled)

            if op in bulk.group.screws_glides:
                checked = qtc.Qt.Checked
            else:
                checked = qtc.Qt.Unchecked
            item.setData(checked, qtc.Qt.CheckStateRole)
            model.setItem(r, 0, item)

        self.__all_widgets['extra_ops'].setModel(model)

    def extra_operations(self):
        """Return the selected operations.

        This method should be called after the dialog
        returns with a qtw.QDialog.Accepted exit code.

        Returns
        -------
        str
            A string in the form expected by the property setter
            gl.PlaneGroup.screws_glides, i.e.,
                "r(#, #, ...), m([#, #], [#, #], ...)"
            NB: the setter should be called with a tuple having
            this as its first element, and bulk.cell_shape as
            the second element.
        """
        items = self.__all_widgets['extra_ops'].model().item

        glides = []
        screws = []

        for i in range(self.n_extra_ops):
            if items(i).checkState() == qtc.Qt.Unchecked:
                continue
            data = items(i).data(qtc.Qt.DisplayRole)
            if 'mirror' in data:       # take direction
                glides.append(data.split('mirror/glide across ')[1])
            elif 'rotation' in data:   # take order
                screws.append(data.split('-fold')[0])
            else:
                raise NotImplementedError(
                    f"Unknown operation {data}. Handling of arbitrary "
                    "symmetry operations is not yet sorted (Issue #15)."
                    )

        txt = []
        if screws:
            txt.append(f"r({', '.join(screws)})")
        if glides:
            txt.append(f"m({', '.join(glides)})")
        return ', '.join(txt)








"""Module calc_section of viperleed.calc.sections.

Defines an enumeration of calculation sections.
"""

__authors__ = (
    'Michele Riva (@michele-riva)',
    'Alexander M. Imre (@amimre)',
    )
__copyright__ = 'Copyright (c) 2019-2024 ViPErLEED developers'
__created__ = '2023-01-23'
__license__ = 'GPLv3+'

from enum import Enum
from itertools import chain

from viperleed.calc.lib.base import readIntRange
from viperleed.calc.lib.base import split_string_range


_ALIASES = {  # Exact match first, then check .startswith
    'INITIALIZATION': ('ini', 'init', 'initialisation'),
    'REFCALC': ('ref', 'fd'),
    'RFACTOR_REFCALC': ('rfac', 'rfactor'),
    'RFACTOR_SUPERPOS': ('rfacsuper',),
    'DELTAS': ('del', 'delta', 'deltaamplitudes'),
    'SEARCH': (),
    'SUPERPOS': ('sup', 'super'),
    'DOMAINS': ('dom', 'domain'),
    'ERRORCALC': ('err', 'error'),
    'FD_OPTIMIZATION': ('opt', 'optimize', 'fdopt'),
    }


class CalcSection(Enum):
    """An enumeration of calculation sections."""
    INITIALIZATION = 0
    REFCALC = 1
    RFACTOR_REFCALC = 11
    RFACTOR_SUPERPOS = 12
    DELTAS = 2
    SEARCH = 3
    SUPERPOS = 31
    DOMAINS = 4
    ERRORCALC = 5
    FD_OPTIMIZATION = 6

    @classmethod
    def _missing_(cls, value):  # Used only with __call__ in EnumMeta!
        """Attempt returning an element from its aliases."""
        if not isinstance(value, str):
            return None
        value = value.lower()
        # See if value is an integer
        try:
            value_int = int(value)
        except ValueError:
            pass
        else:
            try:
                return cls(value_int)
            except ValueError:
                pass

        # Try with the aliases instead. First look for an exact match
        for name, aliases in _ALIASES.items():
            if any(alias == value for alias in aliases):
                return cls[name]
        # Alternatively, see if the value begins with an alias
        for name, aliases in _ALIASES.items():
            if any(value.startswith(alias) for alias in aliases):
                return cls[name]
        return None

    @classmethod
    def sequence_from_string(cls, string_):
        """Return a tuple of sections from a string."""
        # See if string is a single section
        try:
            return (cls(string_),)
        except ValueError:
            pass

        # May be a range: convert it to its integer representation;
        try:
            start, stop = split_string_range(string_)
        except ValueError:
            raise ValueError(
                f'{cls.__name__}: {string_!r} is neither a valid '
                'section nor a valid range of sections.'
                ) from None
        try:
            _as_int_range = f'{cls(start).value}-{cls(stop).value}'
        except ValueError:
            raise ValueError(
                f'{cls.__name__}: Could not interpret one of '
                f'{start!r} and/or {stop!r} as a valid section'
                ) from None
        return tuple(cls(v) for v in readIntRange(_as_int_range))

    @property
    def history_tag(self):
        """Return an identifier suitable for history records."""
        _history_sections = (CalcSection.REFCALC,
                             CalcSection.DELTAS,
                             CalcSection.SEARCH)
        return self.name[0] if self in _history_sections else ''

    @property
    def long_name(self):
        """Return a descriptive name for self."""
        return _LONG_NAMES[self]

    def __lt__(self, other):  # Not sure if we need to implement the others too
        """Return whether self comes before other."""
        try:
            other = self.__class__(other)
        except ValueError:
            return NotImplemented
        return _SECTION_ORDERING[self] < _SECTION_ORDERING[other]


_LONG_NAMES = {
    CalcSection.INITIALIZATION: 'INITIALIZATION',
    CalcSection.REFCALC: 'REFERENCE CALCULATION',
    CalcSection.DELTAS: 'DELTA-AMPLITUDES',
    CalcSection.SEARCH: 'SEARCH',
    CalcSection.RFACTOR_REFCALC: 'R-FACTOR CALCULATION',
    CalcSection.RFACTOR_SUPERPOS: 'R-FACTOR CALCULATION',
    CalcSection.SUPERPOS: 'SUPERPOS',
    CalcSection.ERRORCALC: 'ERROR CALCULATION',
    CalcSection.FD_OPTIMIZATION: 'FULL-DYNAMIC OPTIMIZATION',
    }


_SECTION_ORDERING = {  # Default order of execution
    v: i
    for i, v in enumerate((
        CalcSection.INITIALIZATION,
        CalcSection.REFCALC,
        CalcSection.FD_OPTIMIZATION,
        CalcSection.RFACTOR_REFCALC,
        CalcSection.DELTAS,
        CalcSection.SEARCH,
        CalcSection.SUPERPOS,
        CalcSection.RFACTOR_SUPERPOS,
        CalcSection.DOMAINS,
        CalcSection.ERRORCALC
        ))
    }


_REQUIRED_FILES = {  # Required input files per section
    CalcSection.INITIALIZATION: (
        'IVBEAMS',
        'PARAMETERS',
        'POSCAR',
        'VIBROCC',
        ),
    CalcSection.REFCALC: (
        'BEAMLIST',
        'IVBEAMS',
        'PARAMETERS',
        'PHASESHIFTS',
        'POSCAR',
        'VIBROCC',
        ),
    CalcSection.DELTAS: (
        'BEAMLIST',
        'DISPLACEMENTS',
        'IVBEAMS',
        'PARAMETERS',
        'PHASESHIFTS',
        'POSCAR',
        'VIBROCC',
        ),
    CalcSection.SEARCH: (
        'BEAMLIST',
        'DISPLACEMENTS',
        'EXPBEAMS',
        'IVBEAMS',
        'PARAMETERS',
        'PHASESHIFTS',
        'POSCAR',
        'VIBROCC',
        ),
    CalcSection.RFACTOR_REFCALC: (
        'BEAMLIST',
        'EXPBEAMS',
        'IVBEAMS',
        'PARAMETERS',
        ),
    CalcSection.RFACTOR_SUPERPOS: (
        'BEAMLIST',
        'PARAMETERS',
        'IVBEAMS',
        'EXPBEAMS',
        ),
    CalcSection.SUPERPOS: (
        'BEAMLIST',
        'DISPLACEMENTS',
        'IVBEAMS',
        'PARAMETERS',
        'POSCAR',
        'VIBROCC',
        ),
    CalcSection.ERRORCALC: (
        'BEAMLIST',
        'DISPLACEMENTS',
        'EXPBEAMS',
        'IVBEAMS',
        'PARAMETERS',
        'PHASESHIFTS',
        'POSCAR',
        'VIBROCC',
        ),
    CalcSection.FD_OPTIMIZATION: (
        'BEAMLIST',
        'EXPBEAMS',
        'IVBEAMS',
        'PARAMETERS',
        'PHASESHIFTS',
        'POSCAR',
        'VIBROCC',
        ),
    }

# set of all input files
ALL_INPUT_FILES = set(chain.from_iterable(_REQUIRED_FILES.values()))

# allowed names for the file containing the experimental beams
# files will be used in precedence from left to right
EXPBEAMS_NAMES = ('EXPBEAMS.csv', 'EXPBEAMS')
ALL_INPUT_FILES.update(set(EXPBEAMS_NAMES))



"""ViPErLEED bookkeeper module of package calc."""

__authors__ = (
    'Florian Kraushofer (@fkraushofer)',
    'Alexander M. Imre (@amimre)',
    'Michele Riva (@michele-riva)',
    )
__copyright__ = 'Copyright (c) 2019-2024 ViPErLEED developers'
__created__ = '2020-01-30'
__license__ = 'GPLv3+'

from enum import Enum
import os
from pathlib import Path
import re
import shutil
import time

from viperleed.calc import DEFAULT_HISTORY
from viperleed.calc import DEFAULT_WORK
from viperleed.calc import DEFAULT_WORK_HISTORY
from viperleed.calc import LOG_PREFIX
from viperleed.calc import ORIGINAL_INPUTS_DIR_NAME
from viperleed.calc.sections.calc_section import ALL_INPUT_FILES
from viperleed.cli_base import ViPErLEEDCLI

_CALC_LOG_PREFIXES = (
    LOG_PREFIX,
    'tleedm',   # For backwards compatibility
    )


class BookkeeperMode(Enum):
    DEFAULT = 'default'  # store last run, but do not overwrite POSCAR, VIBROCC
    CONT = 'cont'        # store last run and overwrite POSCAR, VIBROCC from OUT
    DISCARD = 'discard'  # discard previous run as if it never happened


def _translate_timestamp(s):
    """Takes a timestamp YYMMDD-hhmmss and translates it to format DD.MM.YY
    hh:mm:ss"""
    if len(s) != 13:
        print("Error translating timestamp: Invalid length")
        return s
    return "{}.{}.{} {}:{}:{}".format(s[4:6], s[2:4], s[0:2],
                                      s[7:9], s[9:11], s[11:13])


def store_input_files_to_history(root_path, history_path):
    """Finds and copies input files to history.

    Parameters
    ----------
    root_path : pathlike
        Root directory from which to take files. Should be cwd, not ./work.
    history_path : pathlike
        Path to the history directory in which the files should be stored.
    """
    _root_path, _history_path = Path(root_path), Path(history_path)
    original_inputs_path = _root_path / DEFAULT_WORK / ORIGINAL_INPUTS_DIR_NAME
    if original_inputs_path.is_dir():
        input_origin_path = original_inputs_path
    else:
        input_origin_path = _root_path
        print("Could not find directory 'original_inputs' with unaltered "
              "input files. Files will instead be copied from the root "
              "directory.")

    # only files, no dirs
    files_in_dir =  (f for f in Path(input_origin_path).iterdir()
                     if f.is_file())
    files_to_copy = (file for file in files_in_dir
                     if file.name in ALL_INPUT_FILES)
    # copy files to history
    for file in files_to_copy:
        try:
            shutil.copy2(file, _history_path/file.name)
        except OSError as error_msg:
            print(f"Failed to copy file {file} to history: {error_msg}")


def bookkeeper(mode,
               job_name=None,
               history_name=DEFAULT_HISTORY,
               work_history_name=DEFAULT_WORK_HISTORY,):

    # convert mode to enum if necessary
    _mode = BookkeeperMode(mode)

    # get paths for history and workhistory
    history_path = Path(history_name).resolve()
    work_history_path = Path(work_history_name).resolve()
    tensors_path = Path("Tensors").resolve()
    deltas_path = Path("Deltas").resolve()
    out_path = Path("OUT").resolve()

    # make list of stuff to move
    files_to_move = [d for d in os.listdir() if os.path.isdir(d)
              and (d == "OUT" or d == "SUPP")]
    # logs will be saved to history; calc in root, others in SUPP
    logs_to_move = []
    for file in os.listdir():
        if os.path.isfile(file) and file.endswith(".log"):
            if file.startswith(_CALC_LOG_PREFIXES):
                files_to_move.append(file)
            else:
                logs_to_move.append(file)
    # if there's nothing to move, return.
    if len(files_to_move) == 0:
        found = False
        # check workhist folder:
        if work_history_path.is_dir():
            work_history_dirs = [d for d in work_history_path.iterdir() if
                                (work_history_path / d).is_dir()
                                and d.name.startswith("r")
                                and not ("previous" in d.name)]
            if len(work_history_dirs) > 0:
                found = True
        if not found:
            print("Bookkeeper: Found nothing to do. Exiting...")
            return 1
    # check whether history folder is there. If not, make one
    if not history_path.is_dir():
        try:
            os.mkdir(history_name)
        except Exception:
            print("Error creating history folder.")
            raise
    # figure out the number of the tensor
    tensor_number = 0
    if tensors_path.is_dir():
        indlist = []
        rgx = re.compile(r'Tensors_[0-9]{3}\.zip')
        for file in [f for f in os.listdir("Tensors")
                  if ((tensors_path / f).is_file()
                      and rgx.match(f))]:
            match = rgx.match(file)
            if match.span()[1] == 15:  # exact match
                indlist.append(int(match.group(0)[-7:-4]))
        if indlist:
            tensor_number = max(indlist)
    # figure out the number of the run
    dir_list = [d for d in history_path.iterdir()
                if (history_path / d).is_dir()]
    max_nums = {}  # max. job number per tensor number
    rgx = re.compile(r't[0-9]{3}.r[0-9]{3}_')
    for dir in dir_list:
        match = rgx.match(dir.name)
        if match:
            try:
                r_fac_line = int(dir.name[1:4])
                i = int(dir.name[6:9])
                if r_fac_line not in max_nums:
                    max_nums[r_fac_line] = i
                else:
                    max_nums[r_fac_line] = max(max_nums[r_fac_line], i)
            except (ValueError, IndexError):
                pass
    if tensor_number not in max_nums:
        num = 1  # Tensor is new - if discard: delete
        if _mode is BookkeeperMode.DISCARD:
            tensor_file = tensors_path / "Tensors_{tensor_number:03d}.zip"
            delta_file = deltas_path / "Deltas_{tensor_number:03d}.zip"
            for file in (tensor_file, delta_file):
                if os.path.isfile(file):
                    try:
                        os.remove(file)
                    except Exception:
                        print(f"Failed to discard file {file}")
    else:
        num = max_nums[tensor_number] + 1
    # find old timestamp, if possible
    old_log_files = sorted(
        f for f in os.listdir() if os.path.isfile(f) and
        f.endswith(".log") and f.startswith(_CALC_LOG_PREFIXES)
        )
    last_log_lines = []
    if len(old_log_files) > 0:
        old_timestamp = old_log_files[-1][-17:-4]
        try:
            with open(old_log_files[-1], "r") as read_file:
                last_log_lines = read_file.readlines()
        except Exception:
            pass
    else:
        timestamp = time.strftime("%y%m%d-%H%M%S", time.localtime())
        old_timestamp = f"moved-{timestamp}"
    if _mode is not BookkeeperMode.DISCARD:
        # get dirname
        dirname = f"t{tensor_number:03d}.r{num:03d}_{old_timestamp}"
        if job_name is not None:
            dirname += "_" + job_name
        tensor_dir = Path(history_name).resolve() / dirname
        if os.path.isdir(tensor_dir):
            tensor_dir_2 = f"{tensor_dir}_moved-{timestamp}"
            print(f"Error: Target directory {tensor_dir} already exists. Will "
                  "use {tensor_dir_2} instead.")
            tensor_dir = Path(tensor_dir_2).resolve()
            dirname = os.path.basename(tensor_dir)
        try:
            os.mkdir(tensor_dir)
        except Exception:
            print(f"Error: Could not create target directory {tensor_dir}"
                  "\n Stopping...")
            raise
        # copy (or discard) files
        cwd_path = Path(".")
        store_input_files_to_history(cwd_path, tensor_dir)

    # if CONT, check for POSCAR_OUT / VIBROCC_OUT
    # do not complain if not found, since we move previous runs to the history
    # by default
    if _mode is BookkeeperMode.CONT:
        if os.path.isdir("OUT"):
            for file in ("POSCAR", "VIBROCC"):

                files_in_out = sorted(
                    [f for f in out_path.iterdir()
                    if (out_path / f).is_file()
                    and f.name.startswith(f"{file}_OUT_")
                    and "parabola" not in f.name]
                )
                if len(files_in_out) > 0:
                    path = out_path / files_in_out[-1]
                    try:
                        shutil.copy2(path, file)
                    except Exception:
                        print(f"Error: failed to copy {path} as new {file}.")

    # move old stuff
    for file in files_to_move:
        if _mode is not BookkeeperMode.DISCARD:
            try:
                shutil.move(file, tensor_dir / file)
            except Exception:
                print(f"Error: Failed to move {file}.")
        else:  # delete instead
            try:
                shutil.rmtree(file)
            except NotADirectoryError:
                try:
                    os.remove(file)
                except Exception:
                    print(f"Failed to discard file {file}.")
            except Exception:
                print(f"Failed to discard directory {file}.")
    # move log files to SUPP (except for general viperleed-calc log)
    for log_file in logs_to_move:
        if _mode is not BookkeeperMode.DISCARD:
            try:
                hist_supp_path = tensor_dir / 'SUPP'
                shutil.move(log_file, hist_supp_path / log_file)
            except Exception:
                print(f"Error: Failed to move {log_file}.")
        else:   # delete instead
            try:
                os.remove(log_file)
            except Exception:
                print("Failed to discard file {log_file}.")

    # if there is a workhist folder, go through it and move contents as well
    tensor_nums = {tensor_number}

    if work_history_path.is_dir() and _mode is not BookkeeperMode.DISCARD:
        work_hist_prev = [d for d in os.listdir(work_history_name) if
                        os.path.isdir(os.path.join(work_history_name, d))
                        and rgx.match(d) and ("previous" in d)]
        for dir in work_hist_prev:
            try:
                shutil.rmtree(work_history_path / dir)
            except Exception:
                print(f"Failed to delete {dir} directory from "
                      f"{work_history_path}")
        work_history_dirs = [dir for dir in work_history_path.iterdir() if
                        (work_history_path / dir).is_dir()
                        and rgx.match(dir.name)
                        and not ("previous" in dir.name)
                        and old_timestamp in dir.name]
        for dir in work_history_dirs:
            try:
                tensor_num_2 = int(dir.name[1:4])
                search_num = int(dir.name[6:9])
            except (ValueError, IndexError):
                pass
            else:
                if tensor_num_2 not in max_nums:
                    num = 1
                else:
                    num = max_nums[tensor_num_2] + 1
                newname = (f"t{tensor_num_2:03d}.r{num:03d}.{search_num:03d}"
                           + dir.name[9:])
                try:
                    shutil.move(os.path.join(work_history_name, dir),
                                os.path.join(history_name, newname))
                except OSError:
                    print(f"Error: Failed to move {work_history_path / dir}.")
                tensor_nums.add(tensor_num_2)
    if work_history_path.is_dir():
        if (len(list(work_history_path.iterdir())) == 0
            or _mode is BookkeeperMode.DISCARD):
            try:
                shutil.rmtree(work_history_name)
            except OSError as error:
                if _mode is BookkeeperMode.DISCARD:
                    print(f"Failed to discard workhistory folder: {error}")
                else:
                    print(f"Failed to delete empty {work_history_name} "
                          f"directory: {str(error)}")
    if _mode is BookkeeperMode.DISCARD:  # all done
        return 0
    job_nums = []
    for tensor_number in tensor_nums:
        if tensor_number not in max_nums:
            job_nums.append(1)
        else:
            job_nums.append(max_nums[tensor_number] + 1)
    # look for notes file
    notes_name = ""
    notes = ""
    for fn in ("notes", "notes.txt"):
        if os.path.isfile(fn):
            notes_name = fn
            break
    if notes_name:
        try:
            with open(notes_name, 'r') as read_file:
                notes = read_file.read()
        except Exception:
            print(f"Error: Failed to read {notes_name} file.")
    if notes:
        try:
            with open(notes_name, 'w'):
                pass
        except Exception:
            print(f"Error: Failed to clear the {notes_name} file after "
                  "reading.")
    # write history.info
    spacing = 12
    hist = ""
    if os.path.isfile("history.info"):
        hist += "\n\n"
    if tensor_nums == {0}:
        hist += "# TENSORS ".ljust(spacing) + "None\n"
    else:
        hist += "# TENSORS ".ljust(spacing) + str(tensor_nums)[1:-1] + "\n"
    hist += "# JOB ID ".ljust(spacing) + str(job_nums)[1:-1] + "\n"
    if job_name is not None:
        hist += f"# JOB NAME {job_name} \n"
    if len(last_log_lines) > 0:
        # try to read what was executed
        run_info = ""
        i = len(last_log_lines) - 1
        while i > 0:
            if last_log_lines[i].startswith("Executed segments: "):
                run_info = (last_log_lines[i].split("Executed segments: ")[1]
                           .strip())
                break
            i -= 1
        if run_info:
            hist += "# RUN ".ljust(spacing) + run_info + "\n"
        # now try to read final R-factors
        for j in range(i+1, len(last_log_lines)):
            line = last_log_lines[j]
            if line.startswith("Final R"):
                r_fac_line = ""
                if "refcalc" in line:
                    r_fac_line = "# R REF ".ljust(spacing)
                elif "superpos" in line:
                    r_fac_line = "# R SUPER ".ljust(spacing)
                if r_fac_line:
                    hist += r_fac_line + line.split(":", maxsplit=1)[1].strip() + "\n"

    hist += "# TIME ".ljust(spacing) + f"{_translate_timestamp(old_timestamp)} \n"
    hist += "# FOLDER ".ljust(spacing) + f"{dirname} \n"
    hist += f"Notes: {notes}\n"
    hist += "\n###########\n"
    try:
        with open("history.info", "a") as hist_info_file:
            hist_info_file.write(hist)
    except Exception:
        print("Error: Failed to append to history.info file.")
    return 0


class BookkeeperCLI(ViPErLEEDCLI, cli_name='bookkeeper'):
    """The main command-line interface for the bookkeeper utility."""

    def add_parser_arguments(self, parser):
        """Add bookkeeper arguments to parser."""
        super().add_parser_arguments(parser)
        what_next = parser.add_mutually_exclusive_group()
        what_next.add_argument(
            '-c', '--cont',
            help=('overwrite POSCAR and VIBROCC with POSCAR_OUT and '
                  'VIBROCC_OUT from the OUT folder, if present.'),
            action='store_true'
            )
        what_next.add_argument(
            '-d', '--discard',
            help=('discard all results from the last run, as if it had '
                  'not happened, and do not add anything to history or '
                  'history.info. Note that this will not necessarily '
                  'restore modified input files in the main folder.'),
            action='store_true',
            )
        parser.add_argument(
            '-j', '--job-name',
            help=('define a string to be appended to the name of the history '
                  'folder that is created, and is logged in history.info'),
            type=str
            )
        parser.add_argument(
            '--history-name',
            help=('define the name of the history folder that is '
                  f'created/used. Default is {DEFAULT_HISTORY!r}'),
            type=str,
            default=DEFAULT_HISTORY
            )
        parser.add_argument(
            '--work-history-name',
            help=('define the name of the workhistory folder that is '
                  f'created/used. Default is {DEFAULT_WORK_HISTORY!r}'),
            type=str,
            default=DEFAULT_WORK_HISTORY
            )

    def __call__(self, args=None):
        """Call the bookkeeper with command-line args."""
        parsed_args = self.parse_cli_args(args)

        # Select mode
        if parsed_args.cont:
            mode = BookkeeperMode.CONT
        elif parsed_args.discard:
            mode = BookkeeperMode.DISCARD
        else:   # default
            mode = BookkeeperMode.DEFAULT
        return bookkeeper(mode,
                          job_name=parsed_args.job_name,
                          history_name=parsed_args.history_name,
                          work_history_name=parsed_args.work_history_name,)


if __name__ == '__main__':
    BookkeeperCLI.run_as_script()



"""Module beam of viperleed.calc.classes.

Defines the Beam class, a container of information concerning a
diffraction beam.
"""

__authors__ = (
    'Florian Kraushofer (@fkraushofer)',
    'Alexander M. Imre (@amimre)',
    )
__copyright__ = 'Copyright (c) 2019-2024 ViPErLEED developers'
__created__ = '2019-12-03'
__license__ = 'GPLv3+'

import math

import numpy as np
from quicktions import Fraction


class Beam:
    """Class storing properties and data of a diffraction beam.

    Stores h,k, and can store intensity over energy.

    Attributes
    ----------
    hkfrac : tuple (Fraction, Fraction)
        The h and k coordinates of the beam as Fractions
    hk : tuple (float, float)
        The h and k coordinates of the beam as float
    intens : dict {float: float}
        contains {energy: intensity} pairs
    complex_amplitude : dict {float: complex}
        contains {energy: amplitude} pairs
    label : str
        Label for the beam
    """

    def __init__(self, hk, maxdenom=99):
        if all([isinstance(v, Fraction) for v in hk]):
            self.hkfrac = hk
            self.hk = (float(hk[0]), float(hk[1]))
        else:
            self.hk = hk
            self.hkfrac = (Fraction(hk[0]).limit_denominator(maxdenom),
                           Fraction(hk[1]).limit_denominator(maxdenom))
        self.intens = {}
        self.complex_amplitude = {}
        self.label, _ = self.getLabel()

    @property
    def energies(self):
        return np.fromiter(self.intens.keys(), float)
    
    @property
    def intensities(self):
        return np.fromiter(self.intens.values(), float)
    
    def updateIndex(self, hk, maxdenom=99):
        """Keep values but change indices"""
        if all([isinstance(v, Fraction) for v in hk]):
            self.hkfrac = hk
            self.hk = (float(hk[0]), float(hk[1]))
        else:
            self.hk = hk
            self.hkfrac = (Fraction(hk[0]).limit_denominator(maxdenom),
                           Fraction(hk[1]).limit_denominator(maxdenom))
        self.label, _ = self.getLabel()

    def getLabel(self, style="plain", lwidth=-1):
        """
        Returns a string of format '( 1/2 | -1   )', where a,b in '(a|b)'
        are justified to at least the given width. If h or k end up longer,
        the width of the other one will also be changed.

        Parameters
        ----------
        style : str, optional
            Select a different style for the label.

            plain:
                UTF8-compatible, with hyphen '-' as minus, for log
            minus:
                replace hyphen by proper minus sign, '\u2212'
            overbar:
                use latex overbars instead of minus
        lwidth: int, optional
            width of each component of the label; total width = 2*lwidth+3.
            Leave negative to determine automatically

        Returns
        -------
        out : str
            The new label.
        lwidth : int
            The width of the new label

        """
        out = ""
        i = 0
        if lwidth < 0:
            lwidth = 2
            if style == "overbar":
                lwidth = 1
        while i < 2:
            texlen = 0  # added length from tex commands
            if i == 0:
                wl = math.ceil((lwidth - 1) / 2)  # width for numerator
            v = self.hkfrac[i]
            if v.denominator == 1:
                if style != "overbar":
                    s = str(v.numerator).rjust(max(2, wl))   # use more space
                else:
                    if v.numerator < 0:
                        s = ((r"\overline{" + str(abs(v.numerator)) + "}")
                             .rjust(wl + 11))  # 11 chars for the "\overline{}"
                        texlen += 11
                    else:
                        s = str(v.numerator).rjust(wl)
            else:
                if style != "overbar":
                    s = str(v.numerator).rjust(wl)+"/"+str(v.denominator)
                else:
                    if v.numerator < 0:
                        s = ((r"\overline{" + str(abs(v.numerator)) + "/"
                              + str(v.denominator) + "}")
                             .rjust(wl + 11))  # 11 chars for the "\overline{}"
                        texlen += 11
                    else:
                        s = str(v.numerator).rjust(wl)+"/"+str(v.denominator)
            s = s.ljust(lwidth + texlen)
            if len(s) - texlen > lwidth:
                lwidth = len(s) - texlen
                i = -1     # start over
            elif i == 0:
                if style == "overbar":
                    out = "$"
                else:
                    out = ""
                out += "(" + s + "|"
            else:
                out += s + ")"
                if style == "overbar":
                    out += "$"
            i += 1
        if style == "minus":
            out = out.replace("-", "\u2212")
        return out, lwidth

    def normMax(self):
        """
        Normalizes the beam to maximum, i.e. sets the highest value to
        1.0 and rescales the others accordingly.
        """
        if len(self.intens.values()) > 0:
            m = max(self.intens.values())
            if m > 0:
                for en in self.intens:
                    self.intens[en] /= m

    def isEqual(self, beam, eps=1e-4):
        """
        Checks whether the beam is equal to another beam with a given
        tolerance. Returns True or False.
        """
        if (abs(self.hk[0] - beam.hk[0]) < eps
                and abs(self.hk[1] - beam.hk[1]) < eps):
            return True
        return False

    def isEqual_hk(self, hk, eps=1e-4):
        """
        Checks whether the beam hk is equal to a tuple hk with a given
        tolerance. Returns True or False.
        """
        hk = [float(v) for v in hk]  # in case Fractions were passed
        if (abs(self.hk[0] - hk[0]) < eps
                and abs(self.hk[1] - hk[1]) < eps):
            return True
        return False
